"""
Interface for the class that knows where challenges parts are, and can create
new ones.
"""

__all__ = ['BaseBoilerplate']


class BaseBoilerplate:
    """
    A class to know where challenges parts are, and has the ability to create
    new ones.

    All such functionality should be delegated to a concrete subclass of this,
    since they might be different ways to structure code.
    """
    def extract_from_filename(self, filename):
        """Get year, day, and part, from a filename"""
        raise NotImplementedError()

    def get_part_filename(self, year: int, day: int, part: str,
                          relative: bool = False):
        """Get the filename for a part"""
        raise NotImplementedError()

    def get_day_directory(self, year: int, day: int, relative: bool = False):
        """Get the directory for a day"""
        raise NotImplementedError()

    def get_day_input_filename(self, year: int, day: int,
                               relative: bool = False):
        """Get the path for a day's input"""
        raise NotImplementedError()

    def get_year_directory(self, year: int, relative: bool = False):
        """Get the directory for a year"""
        raise NotImplementedError()

    def get_part_module_name(self, year: int, day: int, part: str):
        """Get the module name for a part"""
        raise NotImplementedError()

    def create_part(self, year, day, part):
        """Create a new part"""
        raise NotImplementedError()



from dataclasses import dataclass, field
from typing import Dict

import click

from aox.styling.shortcuts import e_error, e_value


__all__ = ['BaseSummary', 'SummaryRegistry', 'summary_registry']


@dataclass
class SummaryRegistry:
    """
    A registry for all the summary classes, so that they can be used to update
    sections in (usually) a README.

    For every new `BaseSummary` subclass, you should use the
    `@summary_registry.register` decorator, or the `summary_registry.override`
    decorator if you want to override one of them.
    """
    summary_classes: Dict[str, 'BaseSummary'] = field(default_factory=dict)
    """
    The summary classes, by their marker prefix identifier.
    """

    def register(self, summary_class: 'BaseSummary', override=False):
        """
        Decorator to register a summary class, that should have a unique
        `marker_prefix` defined, or `override=True` if you want to override an
        already registered summary class.
        """
        marker_prefix = summary_class.marker_prefix
        if marker_prefix is NotImplemented:
            raise Exception(
                f"{summary_class.__name__} didn't set 'marker_prefix'")
        if not override:
            existing_summary_class = self.summary_classes.get(marker_prefix)
            if existing_summary_class:
                raise Exception(
                    f"{summary_class.__name__} tried to override "
                    f"'{marker_prefix}' that was registered by "
                    f"{existing_summary_class.__name__}")
        self.summary_classes[marker_prefix] = summary_class
        return summary_class

    def override(self, summary_class):
        """Handy decorator to override a previous summary class"""
        return self.register(summary_class, override=True)

    def update_text(self, text, combined_info):
        """
        Update a text's portion with all the summaries specified

        >>> test_summary_registry = SummaryRegistry()
        >>> @test_summary_registry.register
        ... class SummaryA(BaseSummary):
        ...     marker_prefix = 'a'
        ...     # noinspection PyShadowingNames
        ...     def generate(self, combined_info):
        ...         return '\\nA\\n'
        >>> @test_summary_registry.register
        ... class SummaryA(BaseSummary):
        ...     marker_prefix = 'b'
        ...     # noinspection PyShadowingNames
        ...     def generate(self, combined_info):
        ...         return '\\nB\\n'
        >>> @test_summary_registry.register
        ... class SummaryE(BaseSummary):
        ...     marker_prefix = 'e'
        ...     # noinspection PyShadowingNames
        ...     def generate(self, combined_info):
        ...         return '\\nE\\n'
        >>> test_summary_registry.update_text(
        ...     '[//]: # (a-start)\\n\\n[//]: # (a-end)\\n'
        ...     '[//]: # (b-start)\\n\\n[//]: # (b-end)\\n'
        ...     '[//]: # (c-start)\\n\\n[//]: # (c-end)\\n'
        ...     '[//]: # (d-start)\\n\\n[//]: # (d-end)\\n'
        ... , None)
        '[//]: # (a-start)\\nA\\n[//]: # (a-end)\\n[//]:
            # (b-start)\\nB\\n[//]: # (b-end)\\n[//]:
            # (c-start)\\n\\n[//]: # (c-end)\\n[//]:
            # (d-start)\\n\\n[//]: # (d-end)\\n'
        >>> test_summary_registry.update_text(
        ...     '[//]: # (a-start)\\n\\n[//]: # (a-end)\\n'
        ...     '[//]: # (b-start)\\n\\n[//]: # (b?-end)\\n'
        ...     '[//]: # (c-start)\\n\\n[//]: # (c-end)\\n'
        ...     '[//]: # (d-start)\\n\\n[//]: # (d-end)\\n'
        ... , None)
        The ... did you accidentally remove or change it?
        '[//]: # (a-start)\\nA\\n[//]: # (a-end)\\n[//]:
            # (b-start)\\n\\n[//]: # (b?-end)\\n[//]:
            # (c-start)\\n\\n[//]: # (c-end)\\n[//]:
            # (d-start)\\n\\n[//]: # (d-end)\\n'
        """
        updated = text
        for summary_class in self.summary_classes.values():
            # noinspection PyCallingNonCallable
            updated = summary_class().update_text(updated, combined_info)

        return updated


class BaseSummary:
    """
    A base class for defining summaries.

    It can generate and update a section in text.

    To be usable automatically, subclasses need to be registered with a
    `@summary_registry.register` or a `@summary_registry.override` decorator.
    """
    marker_prefix: str = NotImplemented
    """
    A unique marker, to be used when delimiting a section in a piece of text,
    and also to uniquely identify it in `summary_registry`.
    """

    @property
    def start_marker(self):
        """
        The start marker for a section in a piece of text

        >>> # noinspection PyAbstractClass
        >>> class TestSummary(BaseSummary):
        ...     marker_prefix = 'summary'
        >>> TestSummary().start_marker
        '[//]: # (summary-start)'
        """
        return f"[//]: # ({self.marker_prefix}-start)"

    @property
    def end_marker(self):
        """
        The end marker for the a section in a piece of text. It should always
        come after `start_marker`.

        >>> # noinspection PyAbstractClass
        >>> class TestSummary(BaseSummary):
        ...     marker_prefix = 'summary'
        >>> TestSummary().end_marker
        '[//]: # (summary-end)'
        """
        return f"[//]: # ({self.marker_prefix}-end)"

    def generate(self, combined_info):
        """Generate updated section content"""
        raise NotImplementedError()

    def update_text(self, text, combined_info):
        """
        Update text with a new version of content, if the start and end markers
        are present in it.

        >>> # noinspection PyAbstractClass
        >>> class TestSummary(BaseSummary):
        ...     marker_prefix = 'summary'
        ...     # noinspection PyShadowingNames
        ...     def generate(self, combined_info):
        ...         return '\\nYo!\\n'
        >>> TestSummary().update_text(
        ...     '[//]: # (dummy-start)\\n'
        ...     '\\n'
        ...     '[//]: # (dummy-end)'
        ... , None)
        '[//]: # (dummy-start)\\n\\n[//]: # (dummy-end)'
        >>> TestSummary().update_text(
        ...     '[//]: # (summary-start)\\n'
        ...     '\\n'
        ...     '[//]: # (dummy-end)'
        ... , None)
        The ... did you accidentally remove or change it?
        '[//]: # (summary-start)\\n\\n[//]: # (dummy-end)'
        >>> TestSummary().update_text(
        ...     '[//]: # (dummy-start)\\n'
        ...     '\\n'
        ...     '[//]: # (summary-end)'
        ... , None)
        The ... did you accidentally remove or change it?
        '[//]: # (dummy-start)\\n\\n[//]: # (summary-end)'
        >>> TestSummary().update_text(
        ...     '[//]: # (summary-end)\\n'
        ...     '\\n'
        ...     '[//]: # (summary-start)'
        ... , None)
        Markers ... not the other way around
        '[//]: # (summary-end)\\n\\n[//]: # (summary-start)'
        >>> TestSummary().update_text(
        ...     '[//]: # (summary-start)\\n'
        ...     '\\n'
        ...     '[//]: # (summary-end)'
        ... , None)
        '[//]: # (summary-start)\\nYo!\\n[//]: # (summary-end)'
        """
        summary_indexes = self.get_marker_indexes(text)
        if not summary_indexes:
            return text

        summary_text = self.generate(combined_info)

        return self.replace_text(text, summary_indexes, summary_text)

    def replace_text(self, text, indexes, replacement):
        """
        Replace a portion of text between the indexes (non-inclusive)

        >>> # noinspection PyAbstractClass
        >>> class TestSummary(BaseSummary):
        ...     marker_prefix = 'summary'
        >>> TestSummary().replace_text(
        ...     '[//]: # (summary-start)\\n'
        ...     '\\n'
        ...     '[//]: # (summary-end)'
        ... , (23, 25), '\\nYo!\\n')
        '[//]: # (summary-start)\\nYo!\\n[//]: # (summary-end)'
        """
        start_index, end_index = indexes

        return "".join([
            text[:start_index],
            replacement,
            text[end_index:],
        ])

    def get_marker_indexes(self, text):
        """
        Retrieve the start and end marker indexes in a piece of text

        >>> # noinspection PyAbstractClass
        >>> class TestSummary(BaseSummary):
        ...     marker_prefix = 'summary'
        >>> TestSummary().get_marker_indexes('')
        >>> TestSummary().get_marker_indexes('[//]: # (summary-start)')
        The ... did you accidentally remove or change it?
        >>> TestSummary().get_marker_indexes('[//]: # (summary-end)')
        The ... did you accidentally remove or change it?
        >>> TestSummary().get_marker_indexes(
        ...     '[//]: # (summary-end)\\n'
        ...     '\\n'
        ...     '[//]: # (summary-start)'
        ... )
        Markers ... not the other way around
        >>> TestSummary().get_marker_indexes(
        ...     '[//]: # (summary-start)\\n'
        ...     '\\n'
        ...     '[//]: # (summary-end)'
        ... )
        (23, 25)
        """
        if self.start_marker in text:
            start_index = (
                text.index(self.start_marker)
                + len(self.start_marker)
            )
        else:
            start_index = None
        if self.end_marker in text:
            end_index = text.index(self.end_marker)
        else:
            end_index = None

        if start_index is None and end_index is None:
            return None
        elif start_index is None:
            click.echo(
                f"The end marker '{self.end_marker}' was present but start "
                f"marker '{e_error(self.start_marker)}' wasn't - did you "
                f"accidentally remove or change it?")
            return None
        elif end_index is None:
            click.echo(
                f"The start marker '{self.start_marker}' was present but end "
                f"marker '{e_error(self.end_marker)}' wasn't - did you "
                f"accidentally remove or change it?")
            return None

        if end_index < start_index:
            click.echo(
                f"Markers for {self.marker_prefix} where in the "
                f"{e_error('opposite order')}: {e_value(self.start_marker)} "
                f"should be before {e_value(self.end_marker)}, not the other "
                f"way around")
            return None

        return start_index, end_index


summary_registry = SummaryRegistry()



"""
The definition of the command line interface, using `click`.

Not actual work is defined here, but merely how to interact from the console.
All work is delegated to `Controller`.
"""
from typing import Optional

import click

from ..controller.controller import Controller
from aox.settings import settings_proxy

__all__ = ['cli', 'create_cli']

from aox.version import AOX_VERSION_LABEL


def create_cli():
    """Create a CLI instance to run"""
    controller = Controller()

    @click.group(
        invoke_without_command=True,
        help=(
            "A utility tool to manage your code, and display a summary of your "
            "stars, for your Advent of Code submissions\n"
            "\n"
            f"Version: {AOX_VERSION_LABEL}"
        ),
        short_help="Manage code & stars for AOC submissions",
    )
    @click.option('-v', '--version', 'show_version', is_flag=True)
    @click.pass_context
    def aox(ctx, show_version=False):
        if show_version:
            ctx.invoke(version)
            return
        # If we're about to run `init-settings`, don't load the combined info,
        # to avoid any "missing settings" messages.
        if ctx.invoked_subcommand != 'init-settings':
            settings_proxy.ensure_default()
            controller.reload_combined_info()
        if ctx.invoked_subcommand:
            return
        controller.list_years()

    @aox.command(
        help=(
            f"Show current version: {AOX_VERSION_LABEL}"
        ),
        short_help=f"Current version: {AOX_VERSION_LABEL}",
    )
    def version():
        click.echo(AOX_VERSION_LABEL)

    @aox.command(
        help=(
            "Initialise settings for the `aox` utility. Necessary to customise "
            "the behaviour, and to cache stars data"
        ),
        short_help="Initialise local settings",
    )
    def init_settings():
        controller.init_settings()

    @aox.command(
        help=(
            "Dump all the internal data that `aox` knows about, usually for "
            "debugging purposes. The output is JSON, and normally quite big "
            "and verbose."
        ),
        short_help="Dump internal data",
    )
    def dump_data():
        controller.dump_data()

    @aox.group(
        invoke_without_command=True,
        help=(
            "Create, test & run a particular challenge, or do other operations "
            "on it"
        ),
        short_help="Create and Test & run a challenge",
    )
    @click.argument('year', type=int)
    @click.argument('day', type=int)
    @click.argument('part', type=click.Choice(['a', 'b']))
    @click.option('-p', '--path', 'path', type=str)
    @click.option('-f', '--force', 'force', is_flag=True)
    @click.option('--test', '-t', 'filters_texts', multiple=True)
    @click.option('--debug', '-d', 'debug', is_flag=True)
    @click.option('--debug-interval', '-i', 'debug_interval', type=float,
                  default=5)
    @click.pass_context
    def challenge(ctx, year, day, part, path, force, filters_texts, debug,
                  debug_interval):
        if path is not None:
            year, day, part = settings_proxy().challenges_boilerplate\
                .extract_from_filename(path)
            ctx.params['year'], ctx.params['day'], ctx.params['part'] = \
                year, day, part
        if ctx.invoked_subcommand:
            return
        controller.test_and_run_challenge(
            year, day, part, force, filters_texts, debug, debug_interval)

    @challenge.command(
        help=(
            "Fetch or refresh your personalised input for this challenge from "
            "the AOC website"
        ),
        short_help="Fetch the challenge input",
    )
    @click.pass_context
    def refresh_input(ctx):
        params = ctx.parent.params
        controller.refresh_challenge_input(
            params['year'], params['day'], only_if_empty=False)

    @challenge.command(
        name="url",
        help=(
            "Show the URLs to the day and year for this challenge"
        ),
        short_help="Show day and year URLs",
    )
    @click.pass_context
    def show_urls(ctx):
        params = ctx.parent.params
        controller.show_challenge_urls(params['year'], params['day'])

    @challenge.command(
        name="all",
        help=(
            "Test and run the challenge code"
        ),
        short_help="Test and run challenge",
    )
    @click.option('--test', '-t', 'filters_texts', multiple=True)
    @click.option('--debug', '-d', 'debug', is_flag=True)
    @click.option('--debug-interval', '-i', 'debug_interval', type=float,
                  default=5)
    @click.pass_context
    def run_all(ctx, **params):
        params = {
            **ctx.parent.params,
            **params,
        }
        controller.test_and_run_challenge(
            params['year'], params['day'], params['part'], params['force'],
            params['filters_texts'], params['debug'], params['debug_interval'])

    @challenge.command(
        help=(
            "Test the challenge, based on doctests in the file"
        ),
        short_help="Run tests",
    )
    @click.option('--test', '-t', 'filters_texts', multiple=True)
    @click.pass_context
    def test(ctx, **params):
        params = {
            **ctx.parent.params,
            **params,
        }
        controller.test_challenge(
            params['year'], params['day'], params['part'], params['force'],
            params['filters_texts'])

    @challenge.command(
        help=(
            "Run the challenge and print the solution"
        ),
        short_help="Run the challenge",
    )
    @click.option('--debug', '-d', 'debug', is_flag=True)
    @click.option('--debug-interval', '-i', 'debug_interval', type=float,
                  default=5)
    @click.pass_context
    def run(ctx, **params):
        params = {
            **ctx.parent.params,
            **params,
        }
        controller.run_challenge(
            params['year'], params['day'], params['part'], params['force'],
            params['debug'], params['debug_interval'])

    @challenge.command(
        help=(
            "If the challenge offers an interactive play mode, run it. The "
            "challenge needs to have defined a `Challenge.play` method"
        ),
        short_help="Run interactive mode (if defined)",
    )
    @click.pass_context
    def play(ctx):
        params = ctx.parent.params
        controller.play_challenge(
            params['year'], params['day'], params['part'], params['force'])

    @challenge.command(
        help=(
            "Submit your solution to AOC. You can either provide it (eg if it "
            "takes very long to generate) or run the code and submit the result"
        ),
        short_help="Submit the solution to AOC",
    )
    @click.option('--yes', '-y', 'no_prompt', is_flag=True)
    @click.option('--solution', '-s', 'solution')
    @click.pass_context
    def submit(ctx, **params):
        params = {
            **ctx.parent.params,
            **params,
        }
        controller.get_and_submit_challenge_solution(
            params['year'], params['day'], params['part'], params['force'],
            params['no_prompt'], params['solution'])

    @aox.command(
        help=(
            "Add the code boilerplate for a challenge"
        ),
        short_help="Add challenge boilerplate",
    )
    @click.argument('year', type=int)
    @click.argument('day', type=int)
    @click.argument('part', type=click.Choice(['a', 'b']))
    def add(year: int, day: int, part: str):
        controller.add_challenge(year, day, part)

    @aox.command(
        name='list',
        help=(
            "Display a summary of the stars, either for all the years, or a "
            "particular year"
        ),
        short_help="List years/summarise year",
    )
    @click.argument('year', type=int, required=False, default=None)
    def list_years_and_days(year: Optional[int]):
        if year is None:
            controller.list_years()
        else:
            controller.list_days(year)

    @aox.command(
        help=(
            "Fetch all the stars from your AOC account"
        ),
        short_help="Fetch stars data",
    )
    def fetch():
        controller.fetch_account_info()

    @aox.command(
        help=(
            "Update your repo's README with one or more summary formats, "
            "usually displaying stars and links to local code"
        ),
        short_help="Update your README",
    )
    def update_readme():
        controller.update_readme()

    return aox


cli = create_cli()



"""
Gather stars from the AOC site.

The main entry point is `AccountInfo.from_site`
"""

import json
from dataclasses import dataclass, field
from typing import Dict

from aox.settings import settings_proxy
from aox.web import AccountScraper

__all__ = [
    'AccountInfo',
    'AccountYearInfo',
    'AccountDayInfo',
    'AccountPartInfo',
]


@dataclass
class AccountInfo:
    """
    The top level class for account information: it contains references for all
    the years where a challenge part star exists.

    The main entry point is `from_site`.
    """
    username: str
    total_stars: int
    year_infos: Dict[int, 'AccountYearInfo'] = field(default_factory=dict)

    @classmethod
    def from_site(cls):
        """Scrape the site for the stars for each challenge"""
        collected_data = AccountScraper().collect_data()
        return cls.from_collected_data(collected_data)

    @classmethod
    def from_collected_data(cls, collected_data):
        """
        >>> AccountInfo.from_collected_data({
        ...     "username": "Test User", "total_stars": 5, "years": {
        ...         2020: {"year": 2020, "stars": 5, "days": {
        ...             1: 2,
        ...             2: 1,
        ...             3: 0,
        ...         }}}})
        AccountInfo(username='Test User', total_stars=5, year_infos={2020:
            AccountYearInfo(account_info=..., year=2020, stars=5,
                day_infos={1: AccountDayInfo(year_info=..., day=1, stars=2,
                    part_infos={'a':
                        AccountPartInfo(day_info=..., part='a', has_star=True),
                        'b': AccountPartInfo(day_info=..., part='b',
                            has_star=True)}),
                2: AccountDayInfo(year_info=..., day=2, stars=1,
                    part_infos={'a':
                        AccountPartInfo(day_info=..., part='a', has_star=True),
                        'b': AccountPartInfo(day_info=..., part='b',
                            has_star=False)}),
                3: AccountDayInfo(year_info=..., day=3, stars=0,
                    part_infos={'a':
                        AccountPartInfo(day_info=..., part='a', has_star=False),
                        'b': AccountPartInfo(day_info=..., part='b',
                            has_star=False)})})})
        """
        if collected_data is None:
            return None

        username = collected_data['username']
        if not username:
            return None

        total_stars = collected_data['total_stars']

        account_info = cls(username, total_stars)
        account_info.fill_years_from_collected_data(collected_data)

        return account_info

    @classmethod
    def from_cache(cls):
        site_data_path = settings_proxy().site_data_path
        if not site_data_path or not site_data_path.exists():
            return None

        with site_data_path.open() as f:
            serialised = json.load(f)

        return cls.deserialise(serialised)

    @classmethod
    def deserialise(cls, serialised):
        """
        Read the data from JSON. The data are versioned

        >>> AccountInfo.deserialise(AccountInfo.from_collected_data({
        ...     "username": "Test User", "total_stars": 5, "years": {
        ...         2020: {"year": 2020, "stars": 5, "days": {
        ...             1: 2,
        ...             2: 1,
        ...             3: 0,
        ...         }}}}).serialise())
        AccountInfo(username='Test User', total_stars=5, year_infos={2020:
            AccountYearInfo(account_info=..., year=2020, stars=5,
                day_infos={1: AccountDayInfo(year_info=..., day=1, stars=2,
                    part_infos={'a':
                        AccountPartInfo(day_info=..., part='a', has_star=True),
                        'b': AccountPartInfo(day_info=..., part='b',
                            has_star=True)}),
                2: AccountDayInfo(year_info=..., day=2, stars=1,
                    part_infos={'a':
                        AccountPartInfo(day_info=..., part='a', has_star=True),
                        'b': AccountPartInfo(day_info=..., part='b',
                            has_star=False)}),
                3: AccountDayInfo(year_info=..., day=3, stars=0,
                    part_infos={'a':
                        AccountPartInfo(day_info=..., part='a', has_star=False),
                        'b': AccountPartInfo(day_info=..., part='b',
                            has_star=False)})})})
        """
        if serialised is None:
            return None

        version = serialised.get("version", 0)
        if version == 0:
            return cls.deserialise_v0(serialised)
        elif version == 1:
            return cls.deserialise_v1(serialised)
        elif version == 2:
            return cls.deserialise_v2(serialised)
        else:
            raise Exception(f"Not aware of account info version '{version}'")

    @classmethod
    def deserialise_v0(cls, serialised):
        """Read the data from for version 0"""
        account_info = cls(
            username=serialised['user_name'],
            total_stars=serialised['total_stars'],
        )
        account_info.fill_years_from_serialised(serialised['years'], version=0)

        return account_info

    @classmethod
    def deserialise_v1(cls, serialised):
        """Read the data from for version 1"""
        account_info = cls(
            username=serialised['username'],
            total_stars=serialised['total_stars'],
        )
        account_info.fill_years_from_serialised(serialised['years'], version=1)

        return account_info

    @classmethod
    def deserialise_v2(cls, serialised):
        """Read the data from for version 2"""
        account_info = cls(
            username=serialised['username'],
            total_stars=serialised['total_stars'],
        )
        account_info.fill_years_from_serialised(serialised['years'], version=2)

        return account_info

    def fill_years_from_collected_data(self, collected_data):
        """Parse the years information from collected data"""
        for year_data in collected_data['years'].values():
            year_info = AccountYearInfo(
                self, year_data['year'], year_data['stars'])
            year_info.fill_days_from_collected_data(year_data)
            self.year_infos[year_info.year] = year_info

    def fill_years_from_serialised(self, serialised_years, version):
        """Parse the years from JSON. The data are versioned"""
        if version == 0:
            self.fill_years_from_serialised_v0(serialised_years)
        elif version == 1:
            self.fill_years_from_serialised_v1(serialised_years)
        elif version == 2:
            self.fill_years_from_serialised_v2(serialised_years)
        else:
            raise Exception(f"Not aware of account info version '{version}'")

    def fill_years_from_serialised_v0(self, serialised_years):
        """Parse the years from JSON version 0"""
        for year_str, serialised_year in serialised_years.items():
            year = int(year_str)
            year_info = AccountYearInfo.deserialise(
                serialised_year, 0, self)
            self.year_infos[year] = year_info

    def fill_years_from_serialised_v1(self, serialised_years):
        """Parse the years from JSON version 1"""
        for year_str, serialised_year in serialised_years.items():
            year = int(year_str)
            year_info = AccountYearInfo.deserialise(
                {"year": year, **serialised_year}, 1, self)
            self.year_infos[year] = year_info

    def fill_years_from_serialised_v2(self, serialised_years):
        """Parse the years from JSON version 2"""
        for year_str, serialised_year in serialised_years.items():
            year = int(year_str)
            year_info = AccountYearInfo.deserialise(
                {"year": year, **serialised_year}, 2, self)
            self.year_infos[year] = year_info

    def serialise(self):
        """
        Convert to JSON. The data are versioned

        >>> AccountInfo.from_collected_data({
        ...     "username": "Test User", "total_stars": 5, "years": {
        ...         2020: {"year": 2020, "stars": 5, "days": {
        ...             1: 2,
        ...             2: 1,
        ...             3: 0,
        ...         }}}}).serialise()
        {'version': 2, 'username': 'Test User', 'total_stars': 5, 'years':
            {'2020': {'year': 2020, 'stars': 5, 'days':
                {'1': {'year': 2020, 'day': 1, 'stars': 2, 'parts':
                    {'a': {'year': 2020, 'day': 1, 'part': 'a', 'has_star':
                        True},
                    'b': {'year': 2020, 'day': 1, 'part': 'b', 'has_star':
                        True}}},
                '2': {'year': 2020, 'day': 2, 'stars': 1, 'parts':
                    {'a': {'year': 2020, 'day': 2, 'part': 'a', 'has_star':
                        True},
                    'b': {'year': 2020, 'day': 2, 'part': 'b', 'has_star':
                        False}}},
                '3': {'year': 2020, 'day': 3, 'stars': 0, 'parts':
                    {'a': {'year': 2020, 'day': 3, 'part': 'a', 'has_star':
                        False},
                    'b': {'year': 2020, 'day': 3, 'part': 'b', 'has_star':
                        False}}}}}}}
        """
        return {
            "version": 2,
            "username": self.username,
            "total_stars": self.total_stars,
            "years": {
                str(year): year_info.serialise()
                for year, year_info in self.year_infos.items()
            }
        }


@dataclass
class AccountYearInfo:
    """
    Year information, it also contains information for all the days where a
    challenge part star exists.
    """
    account_info: AccountInfo
    year: int
    stars: int
    day_infos: Dict[int, 'AccountDayInfo'] = field(default_factory=dict)

    @classmethod
    def deserialise(cls, serialised, version, account_info):
        """Read the data from JSON. The data are versioned"""
        if version == 0:
            return cls.from_serialised_v0(serialised, account_info)
        elif version == 1:
            return cls.from_serialised_v1(serialised, account_info)
        elif version == 2:
            return cls.from_serialised_v2(serialised, account_info)
        else:
            raise Exception(
                f"Not aware of account year info version '{version}'")

    @classmethod
    def from_serialised_v0(cls, serialised, account_info):
        """Read the data from for version 0"""
        year_info = cls(
            account_info=account_info,
            year=serialised["year"],
            stars=serialised["stars"],
        )
        year_info.fill_days_from_serialised(serialised["days"], version=0)

        return year_info

    @classmethod
    def from_serialised_v1(cls, serialised, account_info):
        """Read the data from for version 1"""
        year_info = cls(
            account_info=account_info,
            year=serialised["year"],
            stars=serialised["stars"],
        )
        year_info.fill_days_from_serialised(serialised["days"], version=1)

        return year_info

    @classmethod
    def from_serialised_v2(cls, serialised, account_info):
        """Read the data from for version 2"""
        year_info = cls(
            account_info=account_info,
            year=serialised["year"],
            stars=serialised["stars"],
        )
        year_info.fill_days_from_serialised(serialised["days"], version=2)

        return year_info

    def fill_days_from_collected_data(self, year_data):
        """Parse the days information from collected data"""
        for day, stars in year_data['days'].items():
            day_info = AccountDayInfo(self, day, stars)
            day_info.fill_parts()
            self.day_infos[day_info.day] = day_info

    def fill_days_from_serialised(self, serialised_days, version):
        """Parse the years from JSON. The data are versioned"""
        if version == 0:
            self.fill_days_from_serialised_v0(serialised_days)
        elif version == 1:
            self.fill_days_from_serialised_v1(serialised_days)
        elif version == 2:
            self.fill_days_from_serialised_v2(serialised_days)
        else:
            raise Exception(
                f"Not aware of account year info version '{version}'")

    def fill_days_from_serialised_v0(self, serialised_days):
        """Read the data from for version 0"""
        for day_str, day_stars in serialised_days.items():
            day = int(day_str)
            day_info = AccountDayInfo.deserialise(
                {"day": day, "stars": day_stars}, 0, self)
            self.day_infos[day] = day_info

    def fill_days_from_serialised_v1(self, serialised_days):
        """Read the data from for version 1"""
        for day_str, serialised_day in serialised_days.items():
            day = int(day_str)
            day_info = AccountDayInfo.deserialise(serialised_day, 1, self)
            self.day_infos[day] = day_info

    def fill_days_from_serialised_v2(self, serialised_days):
        """Read the data from for version 2"""
        for day_str, serialised_day in serialised_days.items():
            day = int(day_str)
            day_info = AccountDayInfo.deserialise(serialised_day, 2, self)
            self.day_infos[day] = day_info

    def serialise(self):
        """Convert to JSON. The data are versioned"""
        return {
            "year": self.year,
            "stars": self.stars,
            "days": {
                str(day): day_info.serialise()
                for day, day_info in self.day_infos.items()
            }
        }


@dataclass
class AccountDayInfo:
    """
    Year information, it also contains information for all the parts where a
    star exists.
    """
    year_info: AccountYearInfo
    day: int
    stars: int
    part_infos: Dict[str, 'AccountPartInfo'] = field(default_factory=dict)

    @classmethod
    def deserialise(cls, serialised, version, year_info):
        """Read the data from JSON. The data are versioned"""
        if version == 0:
            return cls.from_serialised_v0(serialised, year_info)
        elif version == 1:
            return cls.from_serialised_v1(serialised, year_info)
        elif version == 2:
            return cls.from_serialised_v2(serialised, year_info)
        else:
            raise Exception(
                f"Not aware of account day info version '{version}'")

    @classmethod
    def from_serialised_v0(cls, serialised, year_info):
        """Read the data from for version 0"""
        day_info = cls(
            year_info=year_info,
            day=serialised["day"],
            stars=serialised["stars"],
        )
        day_info.fill_parts()

        return day_info

    @classmethod
    def from_serialised_v1(cls, serialised, year_info):
        """Read the data from for version 1"""
        day_info = cls(
            year_info=year_info,
            day=serialised["day"],
            stars=serialised["stars"],
        )
        day_info.fill_parts_from_serialised(serialised["parts"], version=1)

        return day_info

    @classmethod
    def from_serialised_v2(cls, serialised, year_info):
        """Read the data from for version 2"""
        day_info = cls(
            year_info=year_info,
            day=serialised["day"],
            stars=serialised["stars"],
        )
        day_info.fill_parts_from_serialised(serialised["parts"], version=2)

        return day_info

    @property
    def year(self):
        return self.year_info.year

    def fill_parts(self):
        """Fill part infos just from the stars"""
        for part, min_stars in ('a', 1), ('b', 2):
            self.part_infos[part] = AccountPartInfo(
                self, part, self.stars >= min_stars)

    def fill_parts_from_serialised(self, serialised_parts, version):
        """Parse the years from JSON. The data are versioned"""
        if version == 1:
            self.fill_parts_from_serialised_v1(serialised_parts)
        if version == 2:
            self.fill_parts_from_serialised_v2(serialised_parts)
        else:
            raise Exception(
                f"Not aware of account day info version '{version}'")

    def fill_parts_from_serialised_v1(self, serialised_parts):
        """Parse the years from JSON version 1"""
        for part, serialised_part in serialised_parts.items():
            part_info = AccountPartInfo.deserialise(
                serialised_part, 1, self)
            self.part_infos[part] = part_info

    def fill_parts_from_serialised_v2(self, serialised_parts):
        """Parse the years from JSON version 2"""
        for part, serialised_part in serialised_parts.items():
            part_info = AccountPartInfo.deserialise(
                serialised_part, 2, self)
            self.part_infos[part] = part_info

    def serialise(self):
        """Convert to JSON. The data are versioned"""
        return {
            "year": self.year,
            "day": self.day,
            "stars": self.stars,
            "parts": {
                part: part_info.serialise()
                for part, part_info in self.part_infos.items()
            }
        }


@dataclass
class AccountPartInfo:
    """
    Year information, it also contains information for a part.
    """
    day_info: AccountDayInfo
    part: str
    has_star: bool

    @classmethod
    def deserialise(cls, serialised, version, day_info):
        """Read the data from JSON. The data are versioned"""
        if version == 1:
            return cls.deserialise_v1(serialised, day_info)
        elif version == 2:
            return cls.deserialise_v2(serialised, day_info)
        else:
            raise Exception(
                f"Not aware of account part info version '{version}'")

    @classmethod
    def deserialise_v1(cls, serialised, day_info):
        """Parse the years from JSON version 1"""
        return cls(
            day_info=day_info,
            part=serialised["part"],
            has_star=bool(serialised["stars"]),
        )

    @classmethod
    def deserialise_v2(cls, serialised, day_info):
        """Parse the years from JSON version 1"""
        return cls(
            day_info=day_info,
            part=serialised["part"],
            has_star=serialised["has_star"],
        )

    @property
    def year(self):
        return self.day_info.year

    @property
    def day(self):
        return self.day_info.day

    def serialise(self):
        """Convert to JSON. The data are versioned"""
        return {
            "year": self.year,
            "day": self.day,
            "part": self.part,
            "has_star": self.has_star,
        }



"""
All the operations that are provided by the framework can be done via using
`Controller`.
"""

import distutils.dir_util
import json
from dataclasses import dataclass
from doctest import TestResults
from enum import auto
from typing import Optional

import click
from aox.challenge import Debugger

from aox.settings import Settings, settings_proxy
from aox.model import RepoInfo, AccountInfo, CombinedInfo, CombinedPartInfo
from aox.web import WebAoc
from aox.styling.shortcuts import e_warn, e_value, e_success, e_star, e_error, \
    e_unable, e_suggest
from aox.summary import summary_registry
from aox.testing.doctest_enhanced_testmod import testmod_with_filter
from aox.utils import get_current_directory, StringEnum, Timer, \
    has_method_arguments

current_directory = get_current_directory()


@dataclass
class Controller:
    repo_info: Optional[RepoInfo] = None
    account_info: Optional[AccountInfo] = None
    combined_info: Optional[CombinedInfo] = None
    interactive: bool = True

    def init_settings(self, settings_directory=None):
        """Create a new settings directory for the user, if they're missing"""
        if settings_proxy.has() and settings_proxy().path.exists():
            click.echo(
                f"User settings {e_warn('already exist')} at "
                f"{e_value(str(settings_proxy().path))}. Will not overwrite "
                f"them.")
            self.reload_combined_info()
            return False

        self.create_settings(settings_directory=settings_directory)
        self.reload_combined_info()

        return True

    def create_settings(self, settings_directory=None):
        """Create a new settings directory for the user"""
        if settings_directory is None:
            settings_directory = Settings.DEFAULT_SETTINGS_DIRECTORY
        distutils.dir_util.copy_tree(
            Settings.EXAMPLE_SETTINGS_DIRECTORY,
            str(settings_directory))
        settings = Settings.from_settings_directory(settings_directory)
        settings_proxy.set(settings)
        click.echo(
            f"Initialised {e_success('user settings')} at "
            f"{e_value(str(settings_proxy().settings_directory))}! You should "
            f"now edit {e_value(str(settings_proxy().path))} and "
            f"{e_value(str(settings_proxy().sensitive_users_path))}")

    def list_years(self):
        """List all the years that have code or stars"""
        click.echo(
            f"Found {e_success(str(self.combined_info.years_with_code))} years "
            f"with code and "
            f"{e_star(str(self.combined_info.total_stars) + ' stars')}:")
        for year, year_info in sorted(
                self.combined_info.year_infos.items(), reverse=True):
            if not year_info.has_code and not year_info.stars:
                continue
            click.echo(
                f"  * {e_success(str(year))}: "
                f"{e_success(str(year_info.days_with_code))} days with code "
                f"and {e_star(f'{year_info.stars} stars')}")

    STATUS_ICON_MAP = {
        CombinedPartInfo.STATUS_COMPLETE: e_star('*'),
        CombinedPartInfo.STATUS_FAILED: e_error('x'),
        CombinedPartInfo.STATUS_DID_NOT_ATTEMPT: '',
        CombinedPartInfo.STATUS_COULD_NOT_ATTEMPT: e_unable('!', fg='white'),
    }

    def list_days(self, year: int):
        """List all the days for a particular year"""
        year_info = self.combined_info.year_infos.get(year)
        if not year_info or (
                not year_info.has_code and not year_info.stars):
            click.echo(
                f"Could not find {e_error(str(year))} in code nor any stars")
            return False
        if not year_info.has_code:
            click.echo(
                f"Could not find {e_error(str(year))} in code, but found "
                f"{e_star(f'{year_info.stars} stars')}")
            return False
        click.echo(
            f"Found {e_success(str(year_info.days_with_code))} days with code "
            f"in {e_value(str(year))} with "
            f"{e_star(f'{year_info.stars} stars')}:")
        days_string = ', '.join(
            (
                e_success(str(day))
                + self.STATUS_ICON_MAP[day_info.part_infos['a'].status]
                + self.STATUS_ICON_MAP[day_info.part_infos['b'].status]
            )
            for day, day_info
            in sorted(year_info.day_infos.items(), reverse=True)
            if day_info.has_code
        )
        click.echo(f"  * {days_string}")

        return True

    def dump_data(self):
        """Dump the combined data, for diagnostic purposes"""
        if not self.combined_info:
            click.echo(f"There are {e_error('no data loaded')} to dump")
            return False
        print(json.dumps(self.combined_info.serialise(), indent=2))

        return True

    def fetch_account_info(self):
        """Refresh the stars from the AOC website"""
        account_info = AccountInfo.from_site()
        if account_info is None:
            click.echo(f"Could {e_error('not fetch data')}")
            return False

        if settings_proxy().site_data_path:
            with settings_proxy().site_data_path.open('w') as f:
                json.dump(account_info.serialise(), f, indent=2)

        self.update_combined_info(account_info=account_info)
        click.echo(
            f"Fetched data for {e_success(account_info.username)}: "
            f"{e_star(f'{str(account_info.total_stars)} stars')} in "
            f"{e_success(str(len(account_info.year_infos)))} years")

        return True

    def update_readme(self):
        """
        Update README with summaries, presumably because code or stars were
        added.
        """
        readme_path = settings_proxy().readme_path
        if not readme_path:
            return False
        if not self.combined_info.has_site_data:
            click.echo(
                f"Since {e_error('local site data')} are missing the "
                f"README {e_error('cannot be updated')}: run "
                f"{e_suggest('aox fetch')} first")
            return False

        readme_text = readme_path.read_text()

        updated_readme_text = summary_registry.update_text(
            readme_text, self.combined_info)
        if updated_readme_text == readme_text:
            click.echo(f"No need to update {e_success('README')}")
            return False

        readme_path.write_text(updated_readme_text)
        click.echo(f"Updated {e_success('README')} with site data")

        return True

    def refresh_challenge_input(self, year, day, only_if_empty=True):
        """Refresh a challenge's input from the AOC website"""
        input_path = settings_proxy().challenges_boilerplate\
            .get_day_input_filename(year, day)
        if only_if_empty and input_path.exists() and input_path.lstat().st_size:
            return False

        _input = WebAoc().get_input_page(year, day)
        if not _input:
            click.echo(
                f"Could not update input for {e_error(f'{year} {day}')}")
            return False

        if input_path.exists() and _input == input_path.read_text():
            click.echo(
                f"Input did not change for {e_warn(f'{year} {day}')} "
                f"({e_value(f'{len(_input)} bytes')})")
            return False

        input_path.write_text(_input)

        click.echo(
            f"Updated input for {e_success(f'{year} {day}')} "
            f"({e_value(f'{len(_input)} bytes')}) at "
            f"{e_value(str(input_path))}")

        return True

    def add_challenge(self, year: int, day: int, part: str):
        """Add challenge code boilerplate, if it's not already there"""
        if not settings_proxy().challenges_boilerplate\
                .create_part(year, day, part):
            return False
        self.refresh_challenge_input(year=year, day=day)
        part_filename = self.combined_info.get_part(year, day, part).path
        click.echo(
            f"Added challenge {e_success(f'{year} {day} {part.upper()}')} at "
            f"{e_value(str(part_filename))}")
        self.show_challenge_urls(year, day)

        self.update_combined_info(repo_info=RepoInfo.from_roots())

        return True

    def show_challenge_urls(self, year: int, day: int):
        """Show the links to the day and the year"""
        day_info = self.combined_info.get_day(year, day)
        click.echo(
            f"You can see the challenge at {day_info.get_day_url()}, and the "
            f"whole event at {day_info.get_year_url()}")

    def test_and_run_challenge(self, year, day, part, force, filters_texts,
                               debug, debug_interval):
        challenge_instance = self.get_or_create_challenge(
            year, day, part, force)
        if not challenge_instance:
            return False, None, None
        test_results = self.test_challenge(
            year, day, part, force, filters_texts)
        _, solution = self.run_challenge(
            year, day, part, force, debug, debug_interval)

        return True, test_results, solution

    def test_challenge(self, year, day, part, force, filters_texts):
        challenge_instance = self.get_or_create_challenge(
            year, day, part, force)
        if not challenge_instance:
            return None
        filters_text = " ".join(filters_texts)
        test_modules = challenge_instance.get_test_modules()
        with Timer() as timer:
            modules_and_results = [
                (module,
                 testmod_with_filter(
                     module, optionflags=challenge_instance.optionflags,
                     filters_text=filters_text))
                for module in test_modules
            ]
        results = TestResults(
            attempted=sum(
                result.attempted
                for _, result in modules_and_results
            ),
            failed=sum(
                result.failed
                for _, result in modules_and_results
            ),
        )
        failed_modules = [
            module.__name__
            if module else
            'main'
            for module, result in modules_and_results
            if result.failed
        ]
        if failed_modules:
            click.echo(
                f"{e_error(f'{results.failed}/{results.attempted} tests')} "
                f"in {len(failed_modules)}/{len(test_modules)} modules "
                f"{e_error('failed')} "
                f"in {timer.get_pretty_duration(2)}"
                f": {e_error(', '.join(failed_modules))}")
        else:
            click.echo(
                f"{results.attempted} tests "
                f"in {len(test_modules)} modules "
                f"{e_success('passed')} "
                f"in {timer.get_pretty_duration(2)}")

        return results

    def run_challenge(self, year, day, part, force, debug, debug_interval):
        challenge_instance = self.get_or_create_challenge(
            year, day, part, force)
        if not challenge_instance:
            return False, None
        with Timer() as timer:
            debugger = Debugger(
                enabled=debug, min_report_interval_seconds=debug_interval)
            if has_method_arguments(
                    challenge_instance.default_solve, "debugger"):
                solution = challenge_instance.default_solve(debugger=debugger)
            else:
                solution = challenge_instance.default_solve(debug=debugger)
        if solution is None:
            styled_solution = e_error(str(solution))
        else:
            styled_solution = e_value(str(solution))
        click.echo(
            f"Solution: {styled_solution}"
            f" (in {timer.get_pretty_duration(2)})")

        return True, solution

    def play_challenge(self, year, day, part, force):
        challenge_instance = self.get_or_create_challenge(
            year, day, part, force)
        if not challenge_instance:
            return False, None
        result = challenge_instance.play()

        return True, result

    def get_and_submit_challenge_solution(self, year, day, part, force,
                                          no_prompt, solution):
        attempted, submitted, success = False, False, False

        if not WebAoc().is_configured():
            click.echo(
                f"You haven't set {e_error('AOC_SESSION_ID')} in "
                f"{e_error('user_settings.py')}")
            return attempted, submitted, success

        solution = self.get_solution(
            year, day, part, force, no_prompt, solution)
        if solution is None:
            return attempted, submitted, success

        return self.submit_challenge_solution(
            year, day, part, no_prompt, solution)

    def submit_challenge_solution(self, year, day, part, no_prompt, solution):
        attempted, submitted, success = False, False, False

        part_info = self.combined_info\
            .get_part(year, day, part)
        if not part_info:
            return attempted, submitted, success

        attempted = True
        answer_page = WebAoc().submit_solution(year, day, part, solution)
        if not answer_page:
            return attempted, submitted, success

        submitted = True
        message = answer_page.article.text
        is_final_part = part_info.is_final_part
        success, result = self.get_submission_result(is_final_part, message)
        self.echo_submission_result(solution, message, result)

        if success:
            self.fetch_and_update_readme_if_agreed(no_prompt)

        return attempted, submitted, success

    class SubmissionResult(StringEnum):
        RightAnswer = auto()
        AlreadyCompleted = auto()
        WrongAnswer = auto()
        TooSoon = auto()
        RightFinalAnswer = auto()
        Unknown = auto()

    def echo_submission_result(self, solution, message, result):
        """
        >>> Controller().echo_submission_result(
        ...     "1", "Hi", Controller.SubmissionResult.RightAnswer)
        Congratulations! ...
        >>> Controller().echo_submission_result(
        ...     "1", "Hi", Controller.SubmissionResult.AlreadyCompleted)
        It looks like you have already ...
        >>> Controller().echo_submission_result(
        ...     "1", "Hi", Controller.SubmissionResult.WrongAnswer)
        It looks like 1 was the wrong answer...
        >>> Controller().echo_submission_result(
        ...     "1", "Hi", Controller.SubmissionResult.TooSoon)
        It looks like you need to wait ...
        >>> Controller().echo_submission_result(
        ...     "1", "Hi", Controller.SubmissionResult.RightFinalAnswer)
        Congratulations! ...
        >>> Controller().echo_submission_result(
        ...     "1", "Hi", Controller.SubmissionResult.Unknown)
        It's not clear ...
        """
        if result == self.SubmissionResult.RightAnswer:
            click.echo(
                f"Congratulations! That was {e_success('the right answer')}!\n"
                f"{message}"
            )
        elif result == self.SubmissionResult.AlreadyCompleted:
            click.echo(
                f"It looks like you have {e_warn('already completed it')}:\n"
                f"{message}"
            )
        elif result == self.SubmissionResult.WrongAnswer:
            click.echo(
                f"It looks like {e_value(solution)} was the "
                f"{e_error('wrong answer')}:\n {message}"
            )
        elif result == self.SubmissionResult.TooSoon:
            click.echo(
                f"It looks like you need {e_warn('to wait a bit')}:\n{message}"
            )
        elif result == self.SubmissionResult.RightFinalAnswer:
            click.echo(
                f"Congratulations! "
                f"{e_success('You completed the whole year!')}!\n"
                f"{message}"
            )
        elif result == self.SubmissionResult.Unknown:
            click.echo(
                f"It's not clear {e_warn('what was the response')}:\n{message}"
            )
        else:
            raise Exception(f"Don't know how to convey result '{result}'")

    def get_submission_result(self, is_final_part, message):
        """
        >>> import bs4
        >>> from pathlib import Path
        >>> def get_submission_result_from_html(_is_final_part, html_name):
        ...     return Controller().get_submission_result(
        ...        _is_final_part, bs4.BeautifulSoup(Path('').joinpath(
        ...            'tests', 'test_controller', 'test_controller',
        ...            'submission_replies', html_name).read_text(),
        ...            'html.parser').article.text)
        >>> get_submission_result_from_html(False, 'right_answer.html')
        (True, <SubmissionResult.RightAnswer: '...'>)
        >>> get_submission_result_from_html(False, 'already_completed.html')
        (True, <SubmissionResult.AlreadyCompleted: '...'>)
        >>> get_submission_result_from_html(False, 'wrong_answer.html')
        (False, <SubmissionResult.WrongAnswer: '...'>)
        >>> get_submission_result_from_html(False, 'too_soon.html')
        (False, <SubmissionResult.TooSoon: '...'>)
        >>> get_submission_result_from_html(True, 'right_final_answer.html')
        (True, <SubmissionResult.RightFinalAnswer: '...'>)
        >>> get_submission_result_from_html(False, 'unknown.html')
        (False, <SubmissionResult.Unknown: '...'>)
        """
        if "That's the right answer" in message:
            return True, self.SubmissionResult.RightAnswer
        elif "Did you already complete it" in message:
            return True, self.SubmissionResult.AlreadyCompleted
        elif "That's not the right answer" in message:
            return False, self.SubmissionResult.WrongAnswer
        elif "You gave an answer too recently" in message:
            return False, self.SubmissionResult.TooSoon
        elif is_final_part and 'congratulations' in message.lower():
            return True, self.SubmissionResult.RightFinalAnswer
        else:
            return False, self.SubmissionResult.Unknown

    def fetch_and_update_readme_if_agreed(self, no_prompt):
        if no_prompt:
            fetch_and_update_readme = True
        else:
            fetch_and_update_readme = self.prompt(
                f"Do you want to {e_success('fetch stars')} and "
                f"{e_success('update the README')}?",
                type=bool, default=True, no_prompt_default=False)
        if not fetch_and_update_readme:
            click.echo(
                f"Make sure to do {e_suggest('aox fetch')} and "
                f"{e_suggest('aox update-readme')}")
            return

        click.echo(
            f"Fetching {e_star('stars')} and updating "
            f"{e_success('README')}...")
        self.fetch_account_info()
        self.update_readme()

    def get_solution(self, year, day, part, force, no_prompt, solution):
        part_info = self.combined_info\
            .get_part(year, day, part)
        if not part_info:
            return None
        is_final_part = part_info.is_final_part
        if no_prompt or solution not in (None, ""):
            if is_final_part:
                solve_first = False
                if solution in (None, ""):
                    solution = "1"
            else:
                solve_first = solution in (None, "")
        else:
            solve_first = True

        if solve_first:
            challenge_instance = self.get_or_create_challenge(
                year, day, part, force)
            if not challenge_instance:
                return None

            if is_final_part:
                default_solution = "1"
            else:
                default_solution = ""
            if no_prompt:
                solution = default_solution
            else:
                solution = self.prompt(
                    "Run to get the solution, or enter it manually?",
                    default=default_solution,
                    no_prompt_default=default_solution)
            solve_first = not solution
            if solve_first:
                solution = challenge_instance.default_solve()

        if solution in (None, ""):
            click.echo(f"{e_error('No solution')} was provided")
            return None
        solution = str(solution)

        if not no_prompt:
            old_solution = None
            while old_solution != solution:
                old_solution = solution
                solution = self.prompt(
                    f"Submitting solution {e_value(solution)}",
                    default=solution, no_prompt_default=solution)
        else:
            click.echo(
                f"Submitting solution {e_value(solution)}")
        if solution in (None, ""):
            click.echo(f"{e_error('No solution')} was provided")
            return None

        return solution

    def get_or_create_challenge(self, year, day, part, force):
        challenge_instance = self.combined_info\
            .get_challenge_instance(year, day, part)
        if not challenge_instance:
            if not self.add_challenge_if_agreed(year, day, part, force):
                return
            challenge_instance = self.combined_info \
                .get_challenge_instance(year, day, part)
        return challenge_instance

    def add_challenge_if_agreed(self, year: int, day: int, part: str,
                                force: bool):
        if not force:
            should_create_challenge = self.prompt(
                f"Do you want to create challenge "
                f"{e_value(f'{year} {day} {part.upper()}')}?",
                type=bool, default=True, no_prompt_default=False)
        else:
            should_create_challenge = True
        if not should_create_challenge:
            return False
        self.add_challenge(year=year, day=day, part=part)

        return True

    def reload_combined_info(self):
        """Reload all data from the disk"""
        self.update_combined_info(
            repo_info=RepoInfo.from_roots(),
            account_info=AccountInfo.from_cache(),
        )

    def update_combined_info(self, repo_info=None, account_info=None):
        """
        Update the combined info, presumably after updating either pieces of
        data.
        """
        if repo_info is not None:
            self.repo_info = repo_info
        if account_info is not None:
            self.account_info = account_info
        self.combined_info = CombinedInfo.from_repo_and_account_infos(
            repo_info=self.repo_info,
            account_info=self.account_info,
        )

    def prompt(self, text, no_prompt_default, **kwargs):
        if not self.interactive:
            return no_prompt_default
        return click.prompt(text, **kwargs)



from aox.model import CombinedInfo
from aox.summary.base_summary import BaseSummary, summary_registry


__all__ = ['EventSummary']


@summary_registry.register
class EventSummary(BaseSummary):
    """
    Show the stars for every year, in a horizontal table (columns are years)
    """
    marker_prefix = "event-summary"

    def generate(self, combined_info: CombinedInfo):
        headers = (
            ['Total']
            + list(map(str, sorted((
                year
                for year, year_info in combined_info.year_infos.items()
                if year_info.has_code or year_info.stars
            ), reverse=True)))
        )
        return "\n\n{}\n{}\n{}\n\n".format(
            f"| {' | '.join(headers)} |",
            f"| {' | '.join(['---'] * len(headers))} |",
            "| {} |".format(" | ".join(
                self.get_header(combined_info, header)
                for header in headers
            )),
        )

    def get_header(self, combined_info: CombinedInfo, header):
        if header == 'Total':
            stars = combined_info.total_stars
        else:
            stars = combined_info.year_infos[int(header)].stars

        if header != 'Total' and stars == 50:
            emojis = ':star: :star:'
        else:
            emojis = ':star:'

        return f"{stars} {emojis}"



"""
Provides the interface and an implementation for filtering doctests via a
user-defined filtering expression.

A concrete implementation of `DocTestFilterParser` is used by
`FilteringDocTestFinder`.
"""

import re
from abc import ABC
from doctest import DocTest, Example
from typing import Pattern, Optional, List, Tuple


class InvalidTestFilterException(Exception):
    """
    Raised in `DocTestFilterParser.parse_filters` (and all other methods in it)
    to signify an invalid pattern.
    """


class DocTestFilter(ABC):
    """
    Used by `DocTestFinder` to limit which tests should be run. It's
    usually created by a `DocTestFilterParser`, in response to a user-defined
    filter string.
    """
    def matches_test(self, test: DocTest) -> bool:
        """
        Check if a test should be run based only on the name.
        :param test: The `DocTest` object to check against
        :return: Whether it matches
        """
        raise NotImplementedError()

    def matches_example(self, test: DocTest, example: Example,
                        index: int) -> bool:
        """
        Check if a test's example should be run based on the example's index. It
        is assumed that `matches_test` has already been used to match the `test`
        object.
        :param test: The `DocTest` object which contains `example`
        :param example: The `Example` object to check against
        :param index: The 0-based index of the `Example` within `test`
        :return: Whether it matches
        """
        raise NotImplementedError()


class IndexedDocTestFilter(DocTestFilter):
    """
    Filter in two ways:
    * on the full qualified name of the test (eg 'module.Class.method') via a
    regular expression
    * and also optionally via a list of 0-based text index ranges
    """
    def __init__(self, name_regex: Pattern,
                 number_ranges: Optional[List[range]]):
        """
        :param name_regex: A compiled regular expression from `re.compile`
        :param number_ranges: A list objects that supports the `in` operator.
        Normally this will be a list of `range` objects.
        """
        self.name_regex = name_regex
        self.number_ranges = number_ranges

    def matches_test(self, test: DocTest) -> bool:
        """
        >>> IndexedDocTestFilter(re.compile(r"^method$"), []).matches_test(
        ...     DocTest([], {}, "method", "file.py", 0, ""))
        True
        >>> IndexedDocTestFilter(re.compile(r"^method$"), []).matches_test(
        ...     DocTest([], {}, "another_method", "file.py", 0, ""))
        False
        >>> IndexedDocTestFilter(re.compile(r"^.*method$"), []).matches_test(
        ...     DocTest([], {}, "another_method", "file.py", 0, ""))
        True
        """
        return bool(self.name_regex.match(test.name))

    def matches_example(self, test: DocTest, example: Example, index: int
                        ) -> bool:
        """
        >>> a_test = DocTest([], {}, "another_method", "file.py", 0, "")
        >>> IndexedDocTestFilter(re.compile(r"^.*method$"), None)\\
        ...     .matches_example(a_test, Example("", ""), 0)
        True
        >>> IndexedDocTestFilter(re.compile(r"^.*method$"), [
        ...     range(5),
        ... ]).matches_example(a_test, Example("", ""), 0)
        True
        >>> IndexedDocTestFilter(re.compile(r"^.*method$"), [
        ...     range(5),
        ... ]).matches_example(a_test, Example("", ""), 5)
        False
        >>> IndexedDocTestFilter(re.compile(r"^.*method$"), [
        ...     range(5), range(10, 20), [5],
        ... ]).matches_example(a_test, Example("", ""), 5)
        True
        >>> IndexedDocTestFilter(re.compile(r"^.*method$"), [
        ...     range(10), range(20), [5],
        ... ]).matches_example(a_test, Example("", ""), 5)
        True
        """
        if self.number_ranges is None:
            return True
        return any(
            index
            in number_range
            for number_range in self.number_ranges
        )


class DocTestFilterParser(ABC):
    """
    Base class for parsing a filtering expression for selecting specific tests
    to run.

    It is based on the premise that the expression is a list of filters that
    can be individually parsed and collated together.
    """

    def parse_filters(self, filters_text: str) -> List[DocTestFilter]:
        """
        Convert a user-provided filter text to a list of filters
        :param filters_text: The user input
        :return: A list of `DocTestFilter` instances
        """
        return [
            self.parse_filter(filter_text)
            for filter_text in self.split_filter_texts(filters_text)
        ]

    re_split_filter_texts = re.compile(r"\s+")

    def split_filter_texts(self, filters_text: str) -> List[str]:
        """
        Split the user input into multiple filters, to be parsed
        :param filters_text: The user input
        :return: A list of strings that each should be converted to a
        `DocTestFilter`
        >>> DocTestFilterParser().split_filter_texts("")
        []
        >>> DocTestFilterParser().split_filter_texts(" " * 20)
        []
        >>> DocTestFilterParser().split_filter_texts("     abc def      ghi   ")
        ['abc', 'def', 'ghi']
        """
        filters_text = filters_text.strip()
        if not filters_text:
            return []
        return self.re_split_filter_texts.split(filters_text)

    def parse_filter(self, filter_text: str) -> DocTestFilter:
        """
        Convert the text for a single filter to a `DocTestFilter`
        :param filter_text: Part of the user input
        :return: `DocTestFilter`
        """
        raise NotImplementedError()


class IndexedDocTestFilterParser(DocTestFilterParser):
    """
    Since we allow the user to pass a filter for selecting tests, this class
    helps with parsing that text, and converting it to a list `DocTestFilter`s.
    It accepts filters with the following attributes:
    * Each filter must have a name suffix specifier
    * Name specifiers can have '*' inside them to match any sub-string
    * The name can not be empty or only '*'
    * Each filter can also optionally contain a set of indexes or ranges of
    indexes to limit which examples are run inside a test
    Examples of understood patterns:
    * method
    * group_of_*_methods
    * specific.module.*.method
    * methods_with_optional_suffix*
    * method:5
    * method:2-5
    * method:2-
    * method:-5
    * method:-3,5-10,15-
    """
    def parse_filter(self, filter_text: str) -> DocTestFilter:
        test_name_text, line_numbers_text = self.get_filter_parts(filter_text)
        return IndexedDocTestFilter(
            self.parse_test_name(test_name_text),
            self.parse_number_ranges(line_numbers_text),
        )

    def get_filter_parts(self, filter_text: str) -> Tuple[str, str]:
        """
        Split the filter into two parts: the name specifier and the index
        specifier.
        :param filter_text: Part of the user input
        :return: A tuple of the name specifier and the index specifier
        >>> IndexedDocTestFilterParser().get_filter_parts("method")
        ('method', '')
        >>> IndexedDocTestFilterParser().get_filter_parts("method:512,3,-7")
        ('method', '512,3,-7')
        >>> IndexedDocTestFilterParser().get_filter_parts(":512,3,-7")
        ('', '512,3,-7')
        >>> IndexedDocTestFilterParser().get_filter_parts(":")
        ('', '')
        >>> IndexedDocTestFilterParser().get_filter_parts(
        ...     "method:512,3,-7:") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        """
        filter_text = filter_text.strip()
        parts = filter_text.split(':')
        if len(parts) > 2:
            raise InvalidTestFilterException(
                f"A filter should have at most two parts, name and ranges, "
                f"not {len(parts)}: '{filter_text}'")

        if len(parts) == 1:
            test_name_text, = parts
            numbers_text = ''
        else:
            test_name_text, numbers_text = parts

        return test_name_text, numbers_text

    def parse_test_name(self, test_name_text: str) -> Pattern:
        """
        Convert a name speficier to a regex:
        * Escape the specifier so that it matches literally the name
        * Convert '*' to regex '.*' to allow to match substrings
        * The resulting regex should not be equivalent to '^.*$', ie it should
        match literally some part of the method
        :param test_name_text: The name specifier, perhaps containing '*'
        :return: The compiled regex
        >>> def test(pattern):
        ...     regex = IndexedDocTestFilterParser().parse_test_name(pattern)
        ...     # noinspection PyUnresolvedReferences
        ...     return [
        ...         name
        ...         for name in [
        ...             f'{part}.{_type}.{prefix}{method}{suffix}'
        ...             for part in ['part_a', 'part_b']
        ...             for _type in ['TypeA', 'TypeB']
        ...             for method in ['method', 'function']
        ...             for suffix in ['', '_plus']
        ...             for prefix in ['', 'plus_']
        ...         ]
        ...         if regex.match(name)
        ...     ]
        >>> test("method") # doctest: +NORMALIZE_WHITESPACE
        ['part_a.TypeA.method', 'part_a.TypeA.plus_method',
            'part_a.TypeB.method', 'part_a.TypeB.plus_method',
            'part_b.TypeA.method', 'part_b.TypeA.plus_method',
            'part_b.TypeB.method', 'part_b.TypeB.plus_method']
        >>> test("method*") # doctest: +NORMALIZE_WHITESPACE
        ['part_a.TypeA.method', 'part_a.TypeA.plus_method',
            'part_a.TypeA.method_plus', 'part_a.TypeA.plus_method_plus',
            'part_a.TypeB.method', 'part_a.TypeB.plus_method',
            'part_a.TypeB.method_plus', 'part_a.TypeB.plus_method_plus',
            'part_b.TypeA.method', 'part_b.TypeA.plus_method',
            'part_b.TypeA.method_plus', 'part_b.TypeA.plus_method_plus',
            'part_b.TypeB.method', 'part_b.TypeB.plus_method',
            'part_b.TypeB.method_plus', 'part_b.TypeB.plus_method_plus']
        >>> test("part_a.*method") # doctest: +NORMALIZE_WHITESPACE
        ['part_a.TypeA.method', 'part_a.TypeA.plus_method',
            'part_a.TypeB.method', 'part_a.TypeB.plus_method']
        >>> test("part_a.*method*") # doctest: +NORMALIZE_WHITESPACE
        ['part_a.TypeA.method', 'part_a.TypeA.plus_method',
            'part_a.TypeA.method_plus', 'part_a.TypeA.plus_method_plus',
            'part_a.TypeB.method', 'part_a.TypeB.plus_method',
            'part_a.TypeB.method_plus', 'part_a.TypeB.plus_method_plus']
        >>> test("") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        """
        test_name_text = test_name_text.strip()
        if not test_name_text.replace("*", ""):
            raise InvalidTestFilterException(
                f"You need to specify at least some part of the test name")
        parts = re.split(r'\*+', '*' + test_name_text)
        escaped_parts = map(re.escape, parts)
        return re.compile(f"{'.*'.join(escaped_parts)}$")

    def parse_number_ranges(self, number_ranges_text: str) -> List[range]:
        """
        Convert an index specifier to a list of `range`s, to allow matching
        specific examples. An empty specifier returns one range that should
        match any index.
        Examples:
        * '1': Match a specific example
        * '2-5': Match examples 2 to 5, inclusive
        * '-5': Up to 5, inclusive
        * '2-': From 2 to the end, inclusive
        * '-' or '': Match all examples
        * '-3,5-10,20-': A combination of the above
        :param number_ranges_text: The index specifier for a test
        :return: A list of ranges to match the index of examples
        >>> def test(pattern):
        ...     result = IndexedDocTestFilterParser()\\
        ...         .parse_number_ranges(pattern)
        ...     return sorted(set(sum(map(list, result), [])))[:1010]
        >>> test("") # doctest: +ELLIPSIS
        [0, 1, 2, ..., 1000, 1001, ...]
        >>> test("-") # doctest: +ELLIPSIS
        [0, 1, 2, ..., 1000, 1001, ...]
        >>> test("512") # doctest: +ELLIPSIS
        [512]
        >>> test("-512") # doctest: +ELLIPSIS
        [0, 1, 2, ..., 510, 511, 512]
        >>> test("512-") # doctest: +ELLIPSIS
        [512, 513, 514, ..., 1000, 1001, ...]
        >>> test("256-512") # doctest: +ELLIPSIS
        [256, 257, 258, ..., 510, 511, 512]
        >>> test("512-256") # doctest: +ELLIPSIS
        []
        >>> test("256-512-768") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        >>> test("0xf") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        >>> test("5-abc") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        >>> test("1,5,7") # doctest: +ELLIPSIS
        [1, 5, 7]
        >>> test("1,10-20,7") # doctest: +ELLIPSIS
        [1, 7, 10, 11, 12, ..., 18, 19, 20]
        >>> test("1,20-10,7") # doctest: +ELLIPSIS
        [1, 7]
        >>> test("1,20-10,7,10,11") # doctest: +ELLIPSIS
        [1, 7, 10, 11]
        >>> test("1,10-20,-5") # doctest: +ELLIPSIS
        [0, 1, 2, 3, 4, 5, 10, 11, 12, ..., 18, 19, 20]
        >>> test("600-,10-20,-5") # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        [0, 1, 2, 3, 4, 5, 10, 11, 12, ..., 18, 19, 20, 600, 601, 602,
            ..., 1000, 1001, ...]
        >>> test("600-,10-20,,-5") # doctest: +ELLIPSIS
        [0, 1, 2, ..., 1000, 1001, ...]
        >>> test("600-,10-20-40,,-5") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        >>> test("600-,10-0xf,,-5") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        >>> test("600-,10-abc,,-5") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        """
        return [
            self.parse_number_range(number_range_text)
            for number_range_text in number_ranges_text.split(',')
        ]

    def parse_number_range(self, number_range_text: str) -> range:
        """
        Convert a part of an index specifier to a list of `range`s, to allow
        matching specific examples. An empty specifier returns a range that
        should match any index.
        Examples:
        * '1': Match a specific example
        * '2-5': Match examples 2 to 5, inclusive
        * '-5': Up to 5, inclusive
        * '2-': From 2 to the end, inclusive
        * '-' or '': Match all examples
        :param number_range_text: A part of an index specifier
        :return: A range that matches the respective example indexes
        >>> def test(pattern):
        ...     result = IndexedDocTestFilterParser()\\
        ...         .parse_number_range(pattern)
        ...     return sorted(result)[:1010]
        >>> test("") # doctest: +ELLIPSIS
        [0, 1, 2, ..., 1000, 1001, ...]
        >>> test("-") # doctest: +ELLIPSIS
        [0, 1, 2, ..., 1000, 1001, ...]
        >>> test("512") # doctest: +ELLIPSIS
        [512]
        >>> test("-512") # doctest: +ELLIPSIS
        [0, 1, 2, ..., 510, 511, 512]
        >>> test("512-") # doctest: +ELLIPSIS
        [512, 513, 514, ..., 1000, 1001, ...]
        >>> test("256-512") # doctest: +ELLIPSIS
        [256, 257, 258, ..., 510, 511, 512]
        >>> test("512-256") # doctest: +ELLIPSIS
        []
        >>> test("256-512-768") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        >>> test("0xf") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        >>> test("5-abc") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        """
        number_range_text = number_range_text.strip()
        parts = number_range_text.split('-')
        if len(parts) > 2:
            raise InvalidTestFilterException(
                f"Number ranges must be either a single number "
                f"(eg '512'), or a range either without start (eg '-512'), "
                f"without end (eg '512-') or with just start and end "
                f"(eg '256-512), not with more parts: "
                f"'{number_range_text}'")

        if len(parts) == 1:
            number_text, = parts
            if not number_text:
                start = 0
                end = 10000
            else:
                start = end = self.parse_number(number_text)
        else:
            start_number_text, end_number_text = parts
            if not start_number_text and not end_number_text:
                start = 0
                end = 10000
            elif not start_number_text:
                start = 0
                end = self.parse_number(end_number_text)
            elif not end_number_text:
                start = self.parse_number(start_number_text)
                end = 10000
            else:
                start = self.parse_number(start_number_text)
                end = self.parse_number(end_number_text)

        return range(start, end + 1)

    def parse_number(self, number_text: str) -> int:
        """
        A convenient method to raise an appropriate error if the number to parse
        was not a non-negative integer
        :param number_text: A number string to parse
        :return: The parsed number
        >>> IndexedDocTestFilterParser().parse_number("0")
        0
        >>> IndexedDocTestFilterParser().parse_number("512")
        512
        >>> IndexedDocTestFilterParser().parse_number("-4") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        >>> IndexedDocTestFilterParser()\\
        ...     .parse_number("0xf") # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        doctest.InvalidTestFilterException: ...
        """
        try:
            number = int(number_text)
        except ValueError:
            number = None

        if number is None or number < 0:
            raise InvalidTestFilterException(
                f"Line numbers must be positive integers, not "
                f"'{number_text}'")

        return number



__all__ = [
    'join_rows',
    'align_rows',
    'align_divided_rows',
    'simplify_divided_rows',
]


def join_rows(rows, column_separator=' | ', left_pad='| ', right_pad=' |',
              line_separator='\n', align=True):
    """
    Render an ASCII table.

    By default it aligns the columns.

    >>> print(join_rows([])) # doctest: -NORMALIZE_WHITESPACE
    <BLANKLINE>
    >>> print(join_rows([('a', 'b'), ('c', 'd'), ('e', 'f'), ]))
    | a | b |
    | c | d |
    | e | f |
    >>> print(join_rows([
    ...     ('a12', 'b'), ('c', 'd34'), ('e', 'f'),
    ... ])) # doctest: -NORMALIZE_WHITESPACE
    | a12 | b   |
    | c   | d34 |
    | e   | f   |
    >>> print(join_rows([
    ...     ('a12', 'b'), ('c', 'd34'), ('e', 'f'),
    ... ], align=False)) # doctest: -NORMALIZE_WHITESPACE
    | a12 | b |
    | c | d34 |
    | e | f |
    """
    if align:
        rows = align_rows(rows)
    return line_separator.join(
        f"{left_pad}{column_separator.join(row)}{right_pad}"
        for row in rows
    )


def align_rows(rows):
    """
    Visually align rows in the source text of an ASCII table.

    >>> align_rows([]) # doctest: -NORMALIZE_WHITESPACE
    []
    >>> align_rows([
    ...     ('short',), ('short',), ('short',),
    ... ]) # doctest: -NORMALIZE_WHITESPACE
    [('short',), ('short',), ('short',)]
    >>> align_rows([
    ...     ('short',), ('a bit longer',), ('very very long',)
    ... ]) # doctest: -NORMALIZE_WHITESPACE
    [('short         ',), ('a bit longer  ',), ('very very long',)]
    >>> align_rows([
    ...     ('short', 'short'), ('very very long', 'short'),
    ... ]) # doctest: -NORMALIZE_WHITESPACE
    [('short         ', 'short'), ('very very long', 'short')]
    >>> print('\\n'.join(map(' | '.join, align_rows([
    ...     ('short', 'short'), ('very very long', 'short'),
    ... ])))) # doctest: -NORMALIZE_WHITESPACE
    short          | short
    very very long | short
    """
    if not rows:
        return []
    check_rows_have_equal_size(rows)

    column_lengths = tuple(
        max(map(len, column))
        for column in zip(*rows)
    )
    column_formats = tuple(
        f"{{: <{column_length}}}"
        for column_length in column_lengths
    )

    return [
        tuple(
            _format.format(cell)
            for _format, cell in zip(column_formats, row)
        )
        for row in rows
    ]


def simplify_divided_rows(divided_rows, part_separator=' ', align=True):
    """
    Join the parts for each column, effectively simplifying it to a
    ready-to-render ASCII table.

    By default it aligns the parts in each column.

    >>> simplify_divided_rows([
    ...     [('short', 'short'), ('a', 'b')],
    ...     [('very very long', 'short'), ('normal', '')],
    ... ]) # doctest: -NORMALIZE_WHITESPACE
    [('short          short', 'a      b'), ('very very long short', 'normal  ')]
    >>> simplify_divided_rows([
    ...     [('short', 'short'), ('a', 'b')],
    ...     [('very very long', 'short'), ('normal', '')],
    ... ], align=False) # doctest: -NORMALIZE_WHITESPACE
    [('short short', 'a b'), ('very very long short', 'normal ')]
    """
    if align:
        divided_rows = align_divided_rows(divided_rows)
    return [
        tuple(map(part_separator.join, divided_row))
        for divided_row in divided_rows
    ]


def align_divided_rows(divided_rows):
    """
    Same as `align_rows`, but each column is itself divided in columns that need
    to be aligned. This is useful when you want each column to also have its
    parts aligned.

    >>> align_divided_rows([]) # doctest: -NORMALIZE_WHITESPACE
    []
    >>> align_divided_rows([
    ...     (('1-1-a', '1-1-b'), ('1-2-a', '1-2-b')),
    ...     (('2-1-a', '2-1-b'), ('2-2-a', '2-2-b')),
    ... ]) # doctest: -NORMALIZE_WHITESPACE
    [(('1-1-a', '1-1-b'), ('1-2-a', '1-2-b')), \
(('2-1-a', '2-1-b'), ('2-2-a', '2-2-b'))]
    >>> align_divided_rows([
    ...     (('1-1-a-long', '1-1-b'), ('12a', '1-2-b-very-long')),
    ...     (('2-1-a', '2-1-b-long'), ('2-2-a', '2-2-b')),
    ... ]) # doctest: -NORMALIZE_WHITESPACE
    [(('1-1-a-long', '1-1-b     '), ('12a  ', '1-2-b-very-long')), \
(('2-1-a     ', '2-1-b-long'), ('2-2-a', '2-2-b          '))]
    >>> print('\\n'.join(map(
    ...     lambda row: ' | '.join(map(' ~ '.join, row)).rstrip(),
    ...     align_divided_rows([
    ...         (('1-1-a-long', '1-1-b'), ('12a', '1-2-b-very-long')),
    ...         (('2-1-a', '2-1-b-long'), ('2-2-a', '2-2-b')),
    ...     ])))) # doctest: -NORMALIZE_WHITESPACE
    1-1-a-long ~ 1-1-b      | 12a   ~ 1-2-b-very-long
    2-1-a      ~ 2-1-b-long | 2-2-a ~ 2-2-b
    """
    if not divided_rows:
        return []

    check_rows_have_equal_size(divided_rows)

    columns = list(zip(*divided_rows))

    check_columns_have_equal_size(columns)

    aligned_columns = map(align_rows, columns)
    return list(zip(*aligned_columns))


def check_rows_have_equal_size(rows):
    row_sizes = set(map(len, rows))
    if len(row_sizes) != 1:
        raise Exception(
            f"Expected rows of equal size but got multiple sizes: "
            f"{', '.join(map(str, sorted(row_sizes)))}")


def check_columns_have_equal_size(columns):
    uneven_divided_column_sizes = {
        column_index: set(map(len, column))
        for column_index, column in enumerate(columns)
        if len(set(map(len, column))) != 1
    }
    if uneven_divided_column_sizes:
        uneven_columns_message = ', '.join(
            f'{map(str, sorted(column_sizes))} for {column_index}'
            for column_index, column_sizes in uneven_divided_column_sizes
        )
        raise Exception(
            f"Expected columns to be of equal size but got multiple sizes: "
            f"{uneven_columns_message}")



from typing import Callable, Tuple

__all__ = ['get_method_arguments', 'has_method_arguments']

from unittest.mock import NonCallableMock


def get_method_arguments(method: Callable) -> Tuple[str, ...]:
    """
    Extract the arguments from a method
    >>> get_method_arguments(lambda: None)
    ()
    >>> get_method_arguments(lambda foo, bar: None)
    ('foo', 'bar')
    >>> get_method_arguments(lambda foo, bar, debug=False: None)
    ('foo', 'bar', 'debug')
    >>> def function(a, b):
    ...     c = a + b
    ...     return c
    >>> get_method_arguments(function)
    ('a', 'b')
    """
    if isinstance(method, NonCallableMock):
        return ()
    # noinspection PyUnresolvedReferences
    method_code = method.__code__
    return method_code.co_varnames[:method_code.co_argcount]


def has_method_arguments(method: Callable, *arguments: str) -> bool:
    """
    Check if a method has the specified arguments

    >>> has_method_arguments(lambda: None)
    True
    >>> has_method_arguments(lambda: None, 'debug')
    False
    >>> has_method_arguments(lambda foo, bar: None, 'debug')
    False
    >>> has_method_arguments(lambda foo, bar, debug=False: None, 'debug')
    True
    >>> has_method_arguments(lambda foo, bar, debug=False: None, 'debug', 'baz')
    False
    >>> has_method_arguments(lambda foo, bar, debug=False: None, 'debug', 'bar')
    True
    >>> def function(a, b):
    ...     c = a + b
    ...     return c
    >>> has_method_arguments(function, 'a')
    True
    >>> has_method_arguments(function, 'debug')
    False
    """
    method_arguments = get_method_arguments(method)
    return all(
        argument in method_arguments
        for argument in arguments
    )



#!/usr/bin/env python3
from aox.challenge import BaseChallenge


class Challenge(BaseChallenge):
    def solve(self, _input, debugger):
        """
        >>> Challenge().default_solve()
        42
        """


Challenge.main()
challenge = Challenge()



"""
Gather information about the existence of local code for challenges.

The main entry point is `RepoInfo.from_roots`
"""

import datetime
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict

from aox.settings import settings_proxy
from aox.utils import Literal

__all__ = ['RepoInfo', 'RepoYearInfo', 'RepoDayInfo', 'RepoPartInfo']


@dataclass
class RepoInfo:
    """
    The top level class for repo information: it contains references for all the
    years where challenge part code exists.

    The main entry point is `from_roots`.
    """
    has_code: bool
    year_infos: Dict[int, 'RepoYearInfo'] = field(default_factory=dict)

    YEARS = list(range(2015, datetime.datetime.now().year + 1))

    @classmethod
    def from_roots(cls, existing_files=None):
        """
        Create a tree structure, by looking for the expected filenames

        >>> RepoInfo.from_roots(existing_files=[])
        RepoInfo(has_code=False, year_infos={...})
        >>> RepoInfo.from_roots(
        ...     existing_files=[Path('year_2014/day_01/part_a.py')])
        RepoInfo(has_code=False, year_infos={...})
        >>> RepoInfo.from_roots(
        ...     existing_files=[Path('year_2015/day_01/part_a.py')])
        RepoInfo(has_code=True, year_infos={...})
        """
        repo_info = cls(has_code=False)
        repo_info.fill(existing_files)
        return repo_info

    def fill(self, existing_files=None):
        """
        Add `RepoYearInfo` instances (and fill them) for each year present in
        the filenames.
        """
        for year in self.YEARS:
            self.year_infos[year] = RepoYearInfo.from_year(
                year, self, existing_files=existing_files)
        self.has_code = any(
            year_info.has_code
            for year_info in self.year_infos.values()
        )


@dataclass
class RepoYearInfo:
    """
    Year information, it also contains information for all the days where
    challenge part code exists.
    """
    repo_info: 'RepoInfo'
    year: int
    has_code: bool
    path: Path
    day_infos: Dict[int, 'RepoDayInfo'] = field(default_factory=dict)

    DAYS = list(range(1, 26))

    @classmethod
    def from_year(cls, year, repo_info, existing_files=None):
        year_info = cls(
            repo_info=repo_info,
            year=year,
            has_code=False,
            path=settings_proxy().challenges_boilerplate
            .get_year_directory(year),
        )
        year_info.fill(existing_files)
        return year_info

    def fill(self, existing_files=None):
        """
        Add `RepoDayInfo` instances (and fill them) for each day present in the
        filenames.
        """
        for day in self.DAYS:
            self.day_infos[day] = RepoDayInfo.from_day(
                day, self, existing_files=existing_files)
        self.has_code = any(
            day_info.has_code
            for day_info in self.day_infos.values()
        )


@dataclass
class RepoDayInfo:
    """
    Day information, it also contains information for all the parts where
    challenge part code exists.
    """
    year_info: RepoYearInfo
    day: int
    has_code: bool
    path: Path
    part_infos: Dict[str, 'RepoPartInfo'] = field(default_factory=dict)

    PARTS = ['a', 'b']

    @classmethod
    def from_day(cls, day, year_info, existing_files=None):
        day_info = cls(
            year_info=year_info,
            day=day,
            has_code=False,
            path=settings_proxy().challenges_boilerplate.get_day_directory(
                year_info.year, day),
        )
        day_info.fill(existing_files)
        return day_info

    @property
    def year(self):
        return self.year_info.year

    def fill(self, existing_files=None):
        """
        Add `RepoPartInfo` instances (and fill them) for each part present in
        the filenames.
        """
        for part in self.PARTS:
            self.part_infos[part] = RepoPartInfo.from_part(
                part, self, existing_files=existing_files)
        self.has_code = any(
            part_info.has_code
            for part_info in self.part_infos.values()
        )


@dataclass
class RepoPartInfo:
    """Part information."""
    day_info: RepoDayInfo
    part: Literal['a', 'b']
    has_code: bool
    path: Path
    module_name: str

    @property
    def year(self):
        return self.day_info.year

    @property
    def day(self):
        return self.day_info.day

    @classmethod
    def from_part(cls, part, day_info, existing_files=None):
        path = settings_proxy().challenges_boilerplate\
            .get_part_filename(day_info.year, day_info.day, part)
        module_name = settings_proxy().challenges_boilerplate\
            .get_part_module_name(day_info.year, day_info.day, part)
        if existing_files is None:
            has_code = path.exists()
        else:
            has_code = path in existing_files

        return cls(day_info, part, has_code, path, module_name)



def get_testmod_with_filter():
    """
    If testmod has changed so that it can accept `finder`, use the original
    testmod. Otherwise, use the custom one.
    """
    import doctest

    if 'filters_text' in doctest.testmod.__code__.co_varnames[
            :doctest.testmod.__code__.co_argcount]:
        return doctest.testmod
    else:
        from . import doctest_enhanced_testmod
        return doctest_enhanced_testmod.testmod_with_filter


testmod_with_filter = get_testmod_with_filter()



import timeit
from dataclasses import dataclass
from decimal import Decimal
from typing import Optional, Callable, Union

__all__ = ['Timer', 'DummyTimer', 'pretty_duration']


class DummyTimer:
    """
    A dummy timer, useful for testing. It increments in each call by a specified
    amount, not as time passes.
    """
    def __init__(self, start: float = 0, increment: float = 1):
        self.next_value = start
        self.increment = increment

    def __repr__(self) -> str:
        return f"DT({self.next_value}, {self.increment})"

    def __call__(self) -> float:
        """
        >>> DummyTimer()()
        0
        >>> timer = DummyTimer()
        >>> [timer(), timer(), timer()]
        [0, 1, 2]
        >>> timer = DummyTimer(start=2, increment=3)
        >>> [timer(), timer(), timer()]
        [2, 5, 8]
        """
        value = self.next_value
        self.next_value += self.increment
        return value


@dataclass
class Timer:
    """
    A helper class to keep track of time passed since the creation of the
    object.

    It can either be used directly:
        timer = Timer()
        ...
        print(timer.current_duration)
        ...
        timer.finish()
        print(timer.duration)

    Or as a context manager:
        with Timer() as timer:
            ...
            print(timer.current_duration)
            ...
        print(timer.duration)

    You can also either explicity restart the timer:
        timer = Timer()
        ...
        timer.start()

    Or implicitly by using it as a context manager:
        timer = Timer()
        ...
        with timer:
            ...

    >>> Timer(default_timer=DummyTimer())
    Timer(default_timer=DT(1, 1), start=0, end=None)
    >>> timer = Timer(default_timer=DummyTimer())
    >>> [timer.get_current_time(), timer.get_current_time()]
    [1, 2]
    >>> timer.duration
    >>> [timer.current_duration, timer.current_duration]
    [3, 4]
    >>> with timer as timer2:
    ...     timer2 is timer, timer.duration, timer.current_duration
    (True, None, 1)
    >>> timer
    Timer(default_timer=DT(8, 1), start=5, end=7)
    >>> timer.duration, timer.current_duration
    (2, 3)
    >>> with timer as timer2:
    ...     timer2 is timer, timer.duration, timer.current_duration
    (True, None, 1)
    >>> timer
    Timer(default_timer=DT(12, 1), start=9, end=11)
    >>> timer.duration, timer.current_duration
    (2, 3)
    """
    default_timer: Callable[[], float] = timeit.default_timer
    start: float = None
    end: Optional[float] = None

    def __post_init__(self):
        if self.start is None:
            self.start = self.get_current_time()

    def get_current_time(self) -> float:
        """
        >>> timer = Timer(default_timer=DummyTimer())
        >>> [timer.get_current_time(), timer.get_current_time()]
        [1, 2]
        """
        return self.default_timer()

    def begin(self) -> float:
        """
        >>> timer = Timer(default_timer=DummyTimer())
        >>> timer
        Timer(default_timer=DT(1, 1), start=0, end=None)
        >>> timer.begin()
        1
        >>> timer
        Timer(default_timer=DT(2, 1), start=1, end=None)
        >>> timer.finish()
        2
        >>> timer.begin()
        3
        >>> timer
        Timer(default_timer=DT(4, 1), start=3, end=None)
        """
        self.start = self.get_current_time()
        self.end = None
        return self.start

    def finish(self) -> float:
        """
        >>> timer = Timer(default_timer=DummyTimer())
        >>> timer.finish()
        1
        >>> timer
        Timer(default_timer=DT(2, 1), start=0, end=1)
        """
        self.end = self.get_current_time()
        return self.end

    @property
    def duration(self) -> Optional[float]:
        """
        >>> timer = Timer(default_timer=DummyTimer())
        >>> timer.duration
        >>> timer.finish()
        1
        >>> timer.duration
        1
        >>> timer.begin()
        2
        >>> timer.duration
        >>> timer.finish()
        3
        >>> timer.duration
        1
        """
        if self.end is None:
            return None
        return self.end - self.start

    @property
    def pretty_duration(self) -> Optional[str]:
        """
        >>> timer = Timer(default_timer=DummyTimer())
        >>> timer.pretty_duration
        >>> timer.finish()
        1
        >>> timer.pretty_duration
        '1s'
        >>> timer.begin()
        2
        >>> timer.pretty_duration
        >>> timer.finish()
        3
        >>> timer.pretty_duration
        '1s'
        """
        return self.get_pretty_duration()

    def get_pretty_duration(self, digits: Optional[int] = None
                            ) -> Optional[str]:
        duration = self.duration
        if duration is None:
            return None
        return pretty_duration(duration, digits=digits)

    @property
    def current_duration(self) -> float:
        """
        >>> timer = Timer(default_timer=DummyTimer())
        >>> timer.current_duration
        1
        >>> timer.finish()
        2
        >>> timer.duration
        2
        >>> timer.begin()
        3
        >>> timer.current_duration
        1
        >>> timer.finish()
        5
        >>> timer.duration
        2
        """
        return self.get_current_time() - self.start

    @property
    def pretty_current_duration(self) -> str:
        """
        >>> timer = Timer(default_timer=DummyTimer())
        >>> timer.pretty_current_duration
        '1s'
        >>> timer.finish()
        2
        >>> timer.duration
        2
        >>> timer.begin()
        3
        >>> timer.pretty_current_duration
        '1s'
        >>> timer.finish()
        5
        >>> timer.duration
        2
        """
        return self.get_pretty_current_duration()

    def get_pretty_current_duration(self, digits: Optional[int] = None) -> str:
        return pretty_duration(self.current_duration, digits=digits)

    def __enter__(self) -> 'Timer':
        """
        >>> timer = Timer(default_timer=DummyTimer())
        >>> timer.duration, timer.current_duration
        (None, 1)
        >>> with timer as timer2:
        ...     timer2 is timer, timer.duration, timer.current_duration
        (True, None, 1)
        >>> timer.duration, timer.current_duration
        (2, 3)
        """
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()


PRETTY_TIME_LEVELS = [
    # Suffix, factor, next_factor, rounding
    ('s', 60, False),
    ('m', 60, True),
    ('h', 24, True),
    ('d', 7, True),
    ('w', 52, True),
    ('y', None, True),
]


def pretty_duration(duration: float, digits: Optional[int] = None) -> str:
    """
    >>> list(map(pretty_duration, [0, 1, 2.345, 59.234]))
    ['0s', '1s', '2.345s', '59.234s']
    >>> list(map(pretty_duration, [60, 61, 62.345, 3599.234]))
    ['1m0s', '1m1s', '1m2.345s', '59m59.234s']
    >>> list(map(pretty_duration, [3600, 3601, 3662.345, 86399.234]))
    ['1h0m0s', '1h0m1s', '1h1m2.345s', '23h59m59.234s']
    >>> pretty_duration(1 + 60 * (2 + 60 * (3 + 24 * (4 + 7 * (5 + 52 * 6)))))
    '6y5w4d3h2m1s'
    >>> pretty_duration(1 + 60 * (2 + 60 * (3 + 24 * (4 + 7 * (5 + 52 * 999)))))
    '999y5w4d3h2m1s'
    >>> # noinspection PyUnresolvedReferences
    >>> [pretty_duration(x, 1) for x in [59, 59.234, 3599.234, 86399.234]]
    ['59s', '59.2s', '59m59.2s', '23h59m59.2s']
    """
    if duration is None:
        raise ValueError("pretty_duration cannot display a None value")
    if not duration:
        return "0s"
    is_negative = duration < 0
    if is_negative:
        duration *= -1
    remaining = duration
    if digits is not None:
        remaining = round(remaining, digits)
    pretty = ""
    for suffix, next_factor, rounding in PRETTY_TIME_LEVELS:
        if next_factor is not None:
            value = preserve_precision_modulo(remaining, next_factor)
        else:
            value = remaining
        remaining -= value
        if next_factor is not None:
            remaining /= next_factor
        if rounding:
            value = int(value)
        value = remove_unnecessary_precision(value)
        pretty = f"{value}{suffix}{pretty}"
        if not remaining:
            break

    if is_negative:
        pretty = f"-{pretty}"

    return pretty


def preserve_precision_modulo(number: float, divisor: int) -> float:
    """
    >>> preserve_precision_modulo(3599, 60)
    59.0
    >>> preserve_precision_modulo(3599.2, 60)
    59.2
    """
    return float(Decimal(str(number)) % divisor)


def remove_unnecessary_precision(number: float) -> Union[int, float]:
    """
    >>> remove_unnecessary_precision(59)
    59
    >>> remove_unnecessary_precision(59.0)
    59
    >>> remove_unnecessary_precision(59.1)
    59.1
    >>> remove_unnecessary_precision(59.10)
    59.1
    """
    if number == int(number):
        number = int(number)

    return number



"""
Since :mod:`doctest` doesn't allow filtering tests, this provides that
functionality to allow you to select which tests do you want to run.
"""

import doctest
import sys
from typing import Type

from aox.testing.doctest_filter_parsing import DocTestFilterParser, \
    IndexedDocTestFilterParser
from aox.testing.doctest_filtering import FilteringDocTestFinder


def testmod_with_filter(
        m=None, name=None, globs=None, verbose=None, report=True, optionflags=0,
        extraglobs=None, raise_on_error=False, exclude_empty=False,
        parser_class: Type[DocTestFilterParser] = IndexedDocTestFilterParser,
        finder=None, filters_text=""):
    """
    Augment `testmod` with the ability to filter tests via `filters_text`. It
    leverages a version of `testmod` that lets you pass a custom
    `DocTestFinder` instance.

    Optional keyword arg "filters_or_text" specifies either the raw
    user-specified filter text, or the parsed version, in order to limit which
    tests and examples to run. Check the documentation of
    `FilteringDocTestFinder` to see about the syntax.
    """
    if finder is None and filters_text:
        finder = FilteringDocTestFinder(
            exclude_empty=exclude_empty,
            parser_class=parser_class,
            filters_text=filters_text,
        )
    return testmod_with_finder(
        m, name, globs, verbose, report, optionflags, extraglobs,
        raise_on_error, exclude_empty, finder=finder)


def testmod_with_finder(
        m=None, name=None, globs=None, verbose=None, report=True, optionflags=0,
        extraglobs=None, raise_on_error=False, exclude_empty=False,
        finder=None):
    """
    Augment `testmod` with the ability to pass a custom `DocTestFinder`
    instance, that allows for selecting specific tests.

    Optional keyword arg "finder" specifies a finder instance to use, besides
    the default `DocTestFinder`.
    """
    # If no module was given, then use __main__.
    if m is None:
        # DWA - m will still be None if this wasn't invoked from the command
        # line, in which case the following TypeError is about as good an error
        # as we should expect
        m = sys.modules.get('__main__')

    # Check that we were actually given a module.
    import inspect
    if not inspect.ismodule(m):
        raise TypeError("testmod: module required; %r" % (m,))

    # If no name was given, then use the module's name.
    if name is None:
        name = m.__name__

    # Find, parse, and run all tests in the given module.
    if finder is None:
        finder = doctest.DocTestFinder(exclude_empty=exclude_empty)

    if raise_on_error:
        runner = doctest.DebugRunner(verbose=verbose, optionflags=optionflags)
    else:
        runner = doctest.DocTestRunner(verbose=verbose, optionflags=optionflags)

    for test in finder.find(m, name, globs=globs, extraglobs=extraglobs):
        runner.run(test)

    if report:
        runner.summarize()

    if doctest.master is None:
        doctest.master = runner
    else:
        doctest.master.merge(runner)

    return doctest.TestResults(runner.failures, runner.tries)



"""
The default way to structure parts: `year_xxxx/day_xx/part_x.py`
"""

import re
import shutil
from dataclasses import dataclass
from pathlib import Path

import click

from aox.boilerplate.base_boilerplate import BaseBoilerplate
from aox.settings import settings_proxy

from aox.styling.shortcuts import e_warn, e_value, e_error
from aox.utils import get_current_directory

current_directory = get_current_directory()


__all__ = ['DefaultBoilerplate']


@dataclass
class DefaultBoilerplate(BaseBoilerplate):
    """The default way to structure parts `year_xxxx/day_xx/part_x.py`"""
    re_filename = re.compile(r"^(?:.*/)?year_(\d+)/day_(\d+)/part_([ab]).py$")

    example_year_path: Path = current_directory\
        .joinpath('default_boilerplate_example_year')
    example_day_path: Path = example_year_path.joinpath('example_day')
    example_part_path: Path = example_day_path.joinpath('example_part.py')

    def extract_from_filename(self, filename):
        """
        >>> DefaultBoilerplate().extract_from_filename(
        ...     '/home/user/git/my-aoc/year_2020/day_05/part_a.py')
        (2020, 5, 'a')
        >>> DefaultBoilerplate().extract_from_filename(
        ...     'year_2020/day_05/part_a.py')
        (2020, 5, 'a')
        >>> DefaultBoilerplate().extract_from_filename(
        ...     'year_2020/day_15/part_b.py')
        (2020, 15, 'b')
        >>> DefaultBoilerplate().extract_from_filename(
        ...     'year_2020/day_15/part_c.py')
        Traceback (most recent call last):
        ...
        Exception: ...
        >>> DefaultBoilerplate().extract_from_filename(
        ...     'years_2020/day_15/part_a.py')
        Traceback (most recent call last):
        ...
        Exception: ...
        """
        match = self.re_filename.match(filename)
        if not match:
            raise Exception(
                f"Cannot parse filename '{filename}' as a valid challenge part "
                f"filename")

        year_str, day_str, part = match.groups()
        year = int(year_str)
        day = int(day_str)

        return year, day, part

    def get_part_filename(self, year: int, day: int, part: str,
                          relative: bool = False):
        """
        >>> str(DefaultBoilerplate().get_part_filename(2020, 5, 'a', True))
        'year_2020/day_05/part_a.py'
        >>> str(DefaultBoilerplate().get_part_filename(2020, 15, 'a', True))
        'year_2020/day_15/part_a.py'
        """
        day_directory = self.get_day_directory(year, day, relative=relative)
        if day_directory is None:
            return None
        return day_directory.joinpath(f"part_{part}.py")

    def get_day_input_filename(self, year: int, day: int,
                               relative: bool = False):
        """
        >>> str(DefaultBoilerplate().get_day_input_filename(2020, 5, True))
        'year_2020/day_05/input.txt'
        >>> str(DefaultBoilerplate().get_day_input_filename(2020, 15, True))
        'year_2020/day_15/input.txt'
        """
        day_directory = self.get_day_directory(year, day, relative=relative)
        if day_directory is None:
            return None
        return day_directory.joinpath("input.txt")

    def get_day_directory(self, year: int, day: int, relative: bool = False):
        """
        >>> str(DefaultBoilerplate().get_day_directory(2020, 5, True))
        'year_2020/day_05'
        >>> str(DefaultBoilerplate().get_day_directory(2020, 15, True))
        'year_2020/day_15'
        """
        year_directory = self.get_year_directory(year, relative=relative)
        if year_directory is None:
            return None
        return year_directory.joinpath(f"day_{day:0>2}")

    def get_year_directory(self, year: int, relative: bool = False):
        """
        >>> str(DefaultBoilerplate().get_year_directory(2020, True))
        'year_2020'
        """
        if relative:
            base = Path()
        else:
            base = settings_proxy().challenges_root
        if base is None:
            return None
        return base.joinpath(f"year_{year}")

    def get_part_module_name(self, year, day, part):
        """
        >>> DefaultBoilerplate().get_part_module_name(2020, 5, 'a')
        'year_2020.day_05.part_a'
        >>> DefaultBoilerplate().get_part_module_name(2020, 15, 'a')
        'year_2020.day_15.part_a'
        """
        return ".".join(filter(None, [
            settings_proxy().challenges_module_name_root,
            f"year_{year}.day_{day:0>2}.part_{part}",
        ]))

    def create_part(self, year, day, part):
        """Add challenge code boilerplate, if it's not already there"""
        year_path = self.get_year_directory(year)
        day_path = self.get_day_directory(year, day)
        part_path = self.get_part_filename(year, day, part)

        if not any([year_path, day_path, part_path]):
            click.echo(f"You {e_error(f'have not configured')} the root path")
            return False

        if part_path.exists():
            click.echo(
                f"Challenge {e_warn(f'{year} {day} {part.upper()}')} already "
                f"exists at {e_value(str(part_path))}")
            return False

        year_init_path = year_path.joinpath("__init__.py")
        if not year_init_path.exists():
            year_init_path.parent.mkdir(exist_ok=True)
            year_init_path.touch()
        day_init_path = day_path.joinpath("__init__.py")
        if not day_init_path.exists():
            day_init_path.parent.mkdir(exist_ok=True)
            day_init_path.touch()
            part_a_path = self.get_part_filename(year, day, 'a')
            shutil.copy(self.example_part_path, part_a_path)
        day_input_path = day_path.joinpath("input.txt")
        if not day_input_path.exists():
            day_input_path.parent.mkdir(exist_ok=True)
            day_input_path.touch()
        if not part_path.exists():
            shutil.copy(self.example_part_path, part_path)

        return True



from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Optional, TypeVar, Iterable, Callable, List

from aox.settings import settings_proxy
from aox.utils import Timer, pretty_duration
from aox.utils import DummyTimer  # noqa: F401

__all__ = ['Debugger']

T = TypeVar('T')
ReportFormat = Callable[['Debugger', str], str]


@dataclass
class Debugger:
    """
    A debugging helper class, that simplifies outputting debugging information.
    """
    timer: Timer = field(default_factory=Timer)
    """A timer"""
    enabled: bool = True
    """Whether debugging is enabled"""
    step_count: int = 0
    """How many steps have elapsed"""
    step_count_since_last_report: int = 0
    """How many steps have elapsed since the last report?"""
    min_report_interval_seconds: float = 5
    """How much time before two reporting attempts"""
    last_report_time: Optional[float] = None
    """Last time something was reported"""
    extra_report_formats: List[ReportFormat] = field(default_factory=list)
    """A list of nested report formats, to apply successively"""
    indent: str = ""
    """The indent for reporting. Creating a sub-debugger should indent output"""
    indent_increase: str = "  "
    """By how much to increment debugging"""

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.timer.__exit__(exc_tb, exc_val, exc_tb)

    def __bool__(self):
        """
        Convenience method for checking if debugging is enabled

        >>> bool(Debugger())
        True
        >>> bool(Debugger(enabled=False))
        False
        """
        return self.enabled

    def reset(self) -> 'Debugger':
        """
        Reset counting

        >>> debugger = Debugger(timer=Timer(default_timer=DummyTimer()))
        >>> debugger
        Debugger(timer=Timer(default_timer=DT(1, 1), start=0, end=None),
            enabled=True, step_count=0, step_count_since_last_report=0,
            min_report_interval_seconds=5, last_report_time=None,
            extra_report_formats=[], indent='', indent_increase='  ')
        >>> debugger.report()
        Debugger(timer=Timer(default_timer=DT(2, 1), start=0, end=None),
            enabled=True, step_count=0, step_count_since_last_report=0,
            min_report_interval_seconds=5, last_report_time=1,
            extra_report_formats=[], indent='', indent_increase='  ')
        >>> debugger.reset()
        Debugger(timer=Timer(default_timer=DT(3, 1), start=2, end=None),
            enabled=True, step_count=0, step_count_since_last_report=0,
            min_report_interval_seconds=5, last_report_time=None,
            extra_report_formats=[], indent='', indent_increase='  ')
        """
        self.timer.__enter__()
        self.step_count = 0
        self.step_count_since_last_report = 0
        self.last_report_time = None

        return self

    def sub_debugger(self) -> 'Debugger':
        """
        Create a sub-debugger that inherits settings, but starts counting from 0
        time and 0 steps.
        """
        cls = type(self)
        # noinspection PyArgumentList
        return cls(
            enabled=self.enabled,
            min_report_interval_seconds=self.min_report_interval_seconds,
            indent=self.indent + self.indent_increase,
            indent_increase=self.indent_increase,
        )

    @property
    def duration_since_start(self) -> float:
        """How much time has passed since the start"""
        return self.timer.current_duration

    @property
    def pretty_duration_since_start(self) -> str:
        """A pretty rendition of time passed since the start"""
        return self.get_pretty_duration_since_start()

    def get_pretty_duration_since_start(
            self, digits: Optional[int] = 0) -> str:
        """A pretty rendition of time passed since the start"""
        return self.timer.get_pretty_current_duration(digits)

    @property
    def duration_since_last_report(self) -> Optional[float]:
        """
        How much time has passed since the last reporting attempt
        >>> debugger = Debugger(timer=Timer(default_timer=DummyTimer()))
        >>> debugger.duration_since_last_report
        >>> debugger.report()
        Debugger(...)
        >>> debugger.duration_since_last_report
        1
        """
        if self.last_report_time is None:
            return None
        return self.timer.get_current_time() - self.last_report_time

    @property
    def pretty_duration_since_last_report(self) -> Optional[str]:
        """A pretty rendition of time passed since the last reporting attempt"""
        return self.get_pretty_duration_since_last_report()

    def get_pretty_duration_since_last_report(
            self, digits: Optional[int] = 0) -> Optional[str]:
        """A pretty rendition of time passed since the last reporting attempt"""
        if self.last_report_time is None:
            return None
        return pretty_duration(self.last_report_time - self.timer.start, digits)

    @property
    def step_frequency(self) -> float:
        """
        How many steps/second have we performed since the beginning?

        >>> debugger = Debugger(timer=Timer(default_timer=DummyTimer()))
        >>> debugger.step_frequency
        0.0
        >>> debugger.step().step().step()
        Debugger(...)
        >>> debugger.step_frequency
        1.5
        >>> debugger.report().step().step().step()
        Debugger(...)
        >>> debugger.step_frequency
        1.5
        """
        return self.step_count / self.duration_since_start

    @property
    def step_frequency_since_last_report(self) -> float:
        """
        How many steps/second have we performed since the last report (or the
        beginning)?

        >>> debugger = Debugger(timer=Timer(default_timer=DummyTimer()))
        >>> debugger.step_frequency_since_last_report
        0.0
        >>> debugger.step().step().step()
        Debugger(...)
        >>> debugger.step_frequency_since_last_report
        1.5
        >>> debugger.report().step().step().step()
        Debugger(...)
        >>> debugger.step_frequency_since_last_report
        3.0
        """
        duration = self.duration_since_last_report
        if duration is None:
            duration = self.duration_since_start
        return self.step_count_since_last_report / duration

    def should_report(self) -> bool:
        """
        Is the debugger enable, and has enough time passed since the last
        attempt?

        >>> Debugger().should_report()
        True
        >>> Debugger(enabled=False).should_report()
        True
        >>> debugger = Debugger(timer=Timer(default_timer=DummyTimer()))
        >>> debugger.should_report()
        True
        >>> debugger.report()
        Debugger(...)
        >>> [debugger.should_report(), debugger.should_report(),
        ...     debugger.should_report(), debugger.should_report()]
        [False, False, False, False]
        >>> [debugger.should_report(), debugger.should_report(),
        ...     debugger.should_report(), debugger.should_report()]
        [True, True, True, True]
        >>> debugger.report()
        Debugger(...)
        >>> [debugger.should_report(), debugger.should_report(),
        ...     debugger.should_report(), debugger.should_report()]
        [False, False, False, False]
        >>> debugger.should_report()
        True
        """
        if not self.should_report_after_time():
            return False
        return True

    def should_report_after_time(self) -> bool:
        """Has enough time passed since the last attempt?"""
        duration_since_last_report = self.duration_since_last_report
        if duration_since_last_report is None:
            return True
        return (
            duration_since_last_report
            >= self.min_report_interval_seconds
        )

    def step(self, step_count: int = 1) -> 'Debugger':
        """Advance the step counter"""
        self.step_count += step_count
        self.step_count_since_last_report += step_count

        return self

    def stepping(self, items: Iterable[T]) -> Iterable[T]:
        """

        A helper for counting steps, that simply steps once for every item

        >>> debugger = Debugger()
        >>> debugger.step_count
        0
        >>> for index in debugger.stepping(range(3)):
        ...     index, debugger.step_count
        (0, 1)
        (1, 2)
        (2, 3)
        >>> debugger.step_count
        3
        """
        for item in items:
            self.step()
            yield item

    def step_if(self, value: T) -> T:
        """
        A helper for counting steps, that simply steps once for every call if
        the value is truthy, and returns the passed in value.

        >>> debugger = Debugger()
        >>> debugger.step_count
        0
        >>> _value = 0
        >>> while debugger.step_if(_value < 3):
        ...     _value += 1
        ...     _value, debugger.step_count
        (1, 1)
        (2, 2)
        (3, 3)
        >>> debugger.step_count
        3
        """
        if value:
            self.step()
        return value

    @contextmanager
    def adding_extra_report_format(self, report_format: ReportFormat):
        """Temporarily add another report format method"""
        self.extra_report_formats.append(report_format)
        try:
            yield
        finally:
            self.extra_report_formats.remove(report_format)

    def report(self, *args, **kwargs) -> 'Debugger':
        """
        Output a message

        >>> Debugger().report()
        Debugger(...)
        >>> Debugger(enabled=False).report("a", "message")
        Debugger(...)
        >>> Debugger().report("a", "message")
        a message
        Debugger(...)
        >>> Debugger().report("a", "message", sep=",", end="!")
        a,message!Debugger(...)
        >>> Debugger(indent="123").report("a", "message", sep=",", end="!")
        123a,message!Debugger(...)
        >>> Debugger().sub_debugger().report("a", "message")
          a message
        Debugger(...)
        >>> Debugger(indent_increase="??").sub_debugger().report("a", "message")
        ??a message
        Debugger(...)
        """
        if self.enabled and args:
            if self.indent not in (None, ""):
                print(self.indent, end="")
            print(*args, **kwargs)

        self.last_report_time = self.timer.get_current_time()
        self.step_count_since_last_report = 0

        return self

    def default_report(self, message: str = "", **kwargs) -> 'Debugger':
        """
        Output a message with the default format from settings.
        """
        message = self.apply_report_formats(message)
        return self.report(message, **kwargs)

    def apply_report_formats(self, message: str = "") -> str:
        """Apply the default, and any extra report formats, to a message"""
        report_formats = self.extra_report_formats
        for report_format in reversed(report_formats):
            message = report_format(self, message)

        default_debugger_report_format = settings_proxy()\
            .default_debugger_report_format
        if default_debugger_report_format:
            message = default_debugger_report_format(self, message)

        return message

    def report_if(self, *args, **kwargs) -> 'Debugger':
        """
        Output a message, but only if it's appropriate

        >>> debugger = Debugger(timer=Timer(default_timer=DummyTimer()))
        >>> debugger.report_if("a", "message")
        a message
        Debugger(...)
        >>> debugger.report_if("a", "message")
        Debugger(...)
        >>> debugger.report_if("a", "message")
        Debugger(...)
        >>> debugger.report_if("a", "message")
        Debugger(...)
        >>> debugger.report_if("a", "message")
        Debugger(...)
        >>> debugger.report_if("a", "message")
        a message
        Debugger(...)
        """
        if not self.should_report():
            return self
        self.report(*args, **kwargs)

        return self

    def default_report_if(self, message: str = "", **kwargs) -> 'Debugger':
        """
        Output a message with the default format from settings, if it's
        appropriate.
        """
        if not self.should_report():
            return self
        return self.default_report(message, **kwargs)



from dataclasses import dataclass, field, fields
from importlib import import_module
from pathlib import Path
from typing import Optional, Dict, Any, List, Callable, TYPE_CHECKING

import click

from aox.settings import warnings
from aox.settings.warnings import SettingsValidationError
from aox.styling.shortcuts import e_error, e_suggest
from aox.utils import load_module_from_path, get_current_directory, Module
if TYPE_CHECKING:
    from aox.challenge import Debugger  # noqa: F401

__all__ = [
    'InvalidSettingsError',
    'Settings',
]


class InvalidSettingsError(Exception):
    """
    Error signifying that some validation did not pass during settings
    initialisation.
    """
    def __init__(self, message, errors):
        super().__init__(message)
        self.errors = errors


@warnings.register_warnings
@dataclass
class Settings:
    is_missing: bool
    settings_directory: Path
    path: Path
    module: Module
    sensitive_users_path: Path
    aoc_session_id: Optional['str'] = field(
        default=None,
        metadata={
            "module_attribute": "AOC_SESSION_ID",
            "warn": warnings.warn_falsy_attribute,
        },
    )
    challenges_root: Optional[Path] = field(
        default=Path(),
        metadata={
            "module_attribute": "CHALLENGES_ROOT",
            "warn": warnings.warn_missing_path,
        },
    )
    challenges_module_name_root: Optional[str] = field(
        default=None,
        metadata={"module_attribute": "CHALLENGES_MODULE_NAME_ROOT"},
    )
    # noinspection PyUnresolvedReferences
    challenges_boilerplate: 'aox.boilerplate.BaseBoilerplate' = field(  # noqa: F821, E501
        default="aox.boilerplate.DefaultBoilerplate",
        metadata={
            "module_attribute": "CHALLENGES_BOILERPLATE",
            "warn": warnings.error_on(warnings.warn_missing_instance),
            "validate_on_load": True,
        },
    )
    site_data_path: Optional[Path] = field(
        default=None,
        metadata={
            "module_attribute": "SITE_DATA_PATH",
            "warn": warnings.warn_falsy_attribute,
        },
    )
    readme_path: Optional[Path] = field(
        default=Path().joinpath('README.md'),
        metadata={
            "module_attribute": "README_PATH",
            "warn": warnings.warn_missing_path,
        },
    )
    extra_module_imports: List[str] = field(
        default_factory=list,
        metadata={
            "module_attribute": "EXTRA_MODULE_IMPORTS",
        },
    )
    default_debugger_report_format: \
        Optional[Callable[['Debugger', str], str]] = field(
            default=None,
            metadata={
                "module_attribute": "DEFAULT_DEBUGGER_REPORT_FORMAT",
                "warn": warnings.warn_falsy_attribute,
            },
        )
    _warnings: Dict[str, Any] = field(
        default_factory=warnings.get_warnings_for_new_instance)

    DEFAULT_SETTINGS_DIRECTORY = Path('.aox')
    DEFAULT_PATH_NAME = 'user_settings.py'
    DEFAULT_SENSITIVE_USERS_PATH_NAME = 'sensitive_user_settings.py'
    EXAMPLE_SETTINGS_DIRECTORY = \
        get_current_directory().joinpath('.example-aox')

    @classmethod
    def from_default(cls):
        return cls.from_settings_directory(cls.DEFAULT_SETTINGS_DIRECTORY)

    @classmethod
    def from_settings_directory(cls, settings_directory):
        path = settings_directory.joinpath(cls.DEFAULT_PATH_NAME)
        sensitive_users_path = settings_directory\
            .joinpath(cls.DEFAULT_SENSITIVE_USERS_PATH_NAME)
        try:
            settings_module = load_module_from_path(path)
        except (ImportError, FileNotFoundError) as e:
            click.echo(
                f"Could not load {e_error('user_settings.py')} ({e}): using "
                f"default settings - use {e_suggest('aox init-settings')} to "
                f"create your settings file")
            settings_module = None
        return cls.from_settings_module(
            settings_module, settings_directory=settings_directory, path=path,
            sensitive_users_path=sensitive_users_path)

    @classmethod
    def from_settings_module(cls, settings_module, settings_directory, path,
                             sensitive_users_path):
        return cls(
            is_missing=settings_module is None,
            settings_directory=settings_directory,
            path=path,
            module=settings_module,
            sensitive_users_path=sensitive_users_path,
            **{
                _field.name: getattr(
                    settings_module,
                    _field.metadata['module_attribute'],
                )
                for _field in fields(cls)
                if 'module_attribute' in _field.metadata
                and hasattr(
                    settings_module,
                    _field.metadata['module_attribute'],
                )
            },
        )

    def __post_init__(self):
        self.validate()
        self.post_settings_init()

    def validate(self):
        validation_errors = self.get_validation_errors()
        if validation_errors:
            click.echo(
                f"Encountered {e_error('some errors')} while loading settings:"
                f"\n" + "\n".join(
                    f" * {error}"
                    for error in validation_errors
                ))
            raise InvalidSettingsError(
                "Settings were invalid", validation_errors)

    def get_validation_errors(self):
        validation_errors = []
        for _field in fields(self):
            if not _field.metadata.get('validate_on_load', False):
                continue
            try:
                getattr(self, _field.name)
            except SettingsValidationError as e:
                validation_errors.append(e)

        return validation_errors

    def __getattribute__(self, item):
        value = super().__getattribute__(item)
        _warnings = super().__getattribute__('_warnings')
        if item in _warnings:
            warn, module_attribute = _warnings.pop(item)
            value = warn(self, item, module_attribute, value)
            setattr(self, item, value)
        return value

    def post_settings_init(self):
        self.load_extra_module_imports()

    def load_extra_module_imports(self):
        for module_name in self.extra_module_imports:
            import_module(module_name)



import sys

__all__ = ['Literal']


if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing import Any
    from collections import defaultdict
    Literal = defaultdict(lambda: Any)



"""
Combine local and site information about a user's submissions.

The main entry point is `CombinedInfo.from_repo_and_account_infos`.
"""

import itertools
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Optional

import click

from aox.challenge import BaseChallenge
from aox.model.repo_info import RepoInfo
from aox.model.account_info import AccountInfo
from aox.settings import settings_proxy
from aox.styling.shortcuts import e_error
from aox.utils import try_import_module


__all__ = [
    'CombinedInfo',
    'CombinedYearInfo',
    'CombinedDayInfo',
    'CombinedPartInfo',
]

from aox.web import WebAoc


@dataclass
class CombinedInfo:
    """
    The top level class for combined information: it contains references for all
    the years where challenge part code exists.

    The main entry point is `from_repo_and_account_infos`.
    """
    username: Optional[str]
    total_stars: int
    years_with_code: int
    has_site_data: bool
    year_infos: Dict[int, 'CombinedYearInfo'] = field(default_factory=dict)

    @classmethod
    def from_repo_and_account_infos(
            cls, repo_info: RepoInfo, account_info: Optional[AccountInfo]):
        """Main entry point: combine local and remote information"""
        years_with_code = sum(
            1
            for year_info in repo_info.year_infos.values()
            if year_info.has_code
        )
        if account_info is None:
            combined_info = cls(
                username=None,
                total_stars=0,
                years_with_code=years_with_code,
                has_site_data=False,
            )
        else:
            combined_info = cls(
                username=account_info.username,
                total_stars=account_info.total_stars,
                years_with_code=years_with_code,
                has_site_data=True,
            )
            combined_info.amend_years_from_account_info(account_info)
        combined_info.amend_years_from_repo_info(repo_info)

        return combined_info

    def amend_years_from_account_info(self, account_info):
        """
        Called after creating the instance, to populate the years with remote
        information
        """
        for account_year_info in account_info.year_infos.values():
            if account_year_info.year in self.year_infos:
                year_info = self.year_infos[account_year_info.year]
                year_info.stars = account_year_info.stars
            else:
                # noinspection PyTypeChecker
                year_info = CombinedYearInfo(
                    combined_info=self,
                    year=account_year_info.year,
                    stars=account_year_info.stars,
                    path=None,
                    days_with_code=0,
                )
                self.year_infos[account_year_info.year] = year_info
            year_info.amend_days_from_account_year_info(account_year_info)

    def amend_years_from_repo_info(self, repo_info):
        """
        Called after creating the instance, to populate the years with local
        information
        """
        for repo_year_info in repo_info.year_infos.values():
            days_with_code = sum(
                1
                for day_info in repo_year_info.day_infos.values()
                if day_info.has_code
            )
            if repo_year_info.year in self.year_infos:
                year_info = self.year_infos[repo_year_info.year]
                year_info.path = repo_year_info.path
                year_info.days_with_code = days_with_code
            else:
                year_info = CombinedYearInfo(
                    combined_info=self,
                    year=repo_year_info.year,
                    stars=0,
                    path=repo_year_info.path,
                    days_with_code=days_with_code,
                )
                self.year_infos[repo_year_info.year] = year_info
            year_info.amend_days_from_repo_year_info(repo_year_info)

    @property
    def has_code(self):
        return bool(self.years_with_code)

    def get_challenge_instance(self, year, day, part):
        part_info = self.get_part(year, day, part)
        if not part_info:
            return None
        return part_info.get_challenge_instance()

    def get_year(self, year) \
            -> Optional['CombinedYearInfo']:
        year_info = self.year_infos.get(year)
        if not year_info:
            click.echo(
                f"It looks like there is no code for {e_error(str(year))}")
            return None
        return year_info

    def get_day(self, year, day) \
            -> Optional['CombinedDayInfo']:
        year_info = self.get_year(year)
        if not year_info:
            return None
        return year_info.get_day(day)

    def get_part(self, year, day, part) \
            -> Optional['CombinedPartInfo']:
        day_info = self.get_day(year, day)
        if not day_info:
            return None
        return day_info.get_part(part)

    def serialise(self):
        return {
            "version": 1,
            "username": self.username,
            "total_stars": self.total_stars,
            "years_with_code": self.years_with_code,
            "has_site_data": self.has_site_data,
            "years": {
                str(year): year_info.serialise()
                for year, year_info in self.year_infos.items()
            }
        }


@dataclass
class CombinedYearInfo(object):
    """
    Year information, it contains references for all the days where challenge
    part code or stars exist.
    """
    combined_info: CombinedInfo
    year: int
    stars: int
    path: Path
    days_with_code: int
    day_infos: Dict[int, 'CombinedDayInfo'] = field(default_factory=dict)
    counts_by_part_status: Dict[str, int] = field(
        default_factory=lambda:
        CombinedYearInfo.get_initial_counts_by_part_status())

    @classmethod
    def get_initial_counts_by_part_status(cls):
        """Filler for grouping parts by status"""
        return {
            status: 0
            for status in CombinedPartInfo.STATUSES
        }

    @property
    def relative_path(self):
        return self.path.relative_to(settings_proxy().challenges_root)

    def amend_days_from_account_year_info(self, account_year_info):
        """
        Called after creating the instance, to populate the days with remote
        information
        """
        for account_day_info in account_year_info.day_infos.values():
            if account_day_info.day in self.day_infos:
                day_info = self.day_infos[account_day_info.day]
                day_info.stars = account_day_info.stars
            else:
                # noinspection PyTypeChecker
                day_info = CombinedDayInfo(
                    year_info=self,
                    day=account_day_info.day,
                    stars=account_day_info.stars,
                    path=None,
                    parts_with_code=0,
                )
                self.day_infos[account_day_info.day] = day_info
            day_info.amend_parts_from_account_day_info(account_day_info)
        self.update_counts_by_part_status()

    def amend_days_from_repo_year_info(self, repo_year_info):
        """
        Called after creating the instance, to populate the days with local
        information
        """
        for repo_day_info in repo_year_info.day_infos.values():
            parts_with_code = sum(
                1
                for part_info in repo_day_info.part_infos.values()
                if part_info.has_code
            )
            if repo_day_info.day in self.day_infos:
                day_info = self.day_infos[repo_day_info.day]
                day_info.path = repo_day_info.path
                day_info.parts_with_code = parts_with_code
            else:
                day_info = CombinedDayInfo(
                    year_info=self,
                    day=repo_day_info.day,
                    stars=0,
                    path=repo_day_info.path,
                    parts_with_code=parts_with_code,
                )
                self.day_infos[repo_day_info.day] = day_info
            day_info.amend_parts_from_repo_day_info(repo_day_info)
        self.update_counts_by_part_status()

    def update_counts_by_part_status(self):
        """Group parts by status"""
        for day_info in self.day_infos.values():
            for part_info in day_info.part_infos.values():
                part_info.update_status()
        self.counts_by_part_status.update({
            **self.get_initial_counts_by_part_status(),
            **{
                part_status: len(list(items))
                for part_status, items in itertools.groupby(sorted(
                    part_info.status
                    for day_info in self.day_infos.values()
                    for part_info in day_info.part_infos.values()
                ))
            },
        })

    @property
    def has_code(self):
        return bool(self.days_with_code)

    @property
    def has_site_data(self):
        return self.combined_info.has_site_data

    def get_year_url(self) -> str:
        return WebAoc().get_year_url(self.year)

    def get_day(self, day) -> Optional['CombinedDayInfo']:
        day_info = self.day_infos.get(day)
        if not day_info:
            click.echo(
                f"It looks like there is no code for "
                f"{e_error(f'{self.year} day {day}')}")
            return None
        return day_info

    def get_part(self, day, part) -> Optional['CombinedPartInfo']:
        day_info = self.get_day(day)
        if not day_info:
            return None
        return day_info.get_part(part)

    def serialise(self):
        return {
            "year": self.year,
            "stars": self.stars,
            "days_with_code": self.days_with_code,
            "counts_by_part_status": self.counts_by_part_status,
            "days": {
                str(day): day_info.serialise()
                for day, day_info in self.day_infos.items()
            }
        }


@dataclass
class CombinedDayInfo:
    """
    Day information, it contains references for all the parts where challenge
    part code or stars exist.
    """
    year_info: CombinedYearInfo
    day: int
    stars: int
    path: Path
    parts_with_code: int
    part_infos: Dict[str, 'CombinedPartInfo'] = field(default_factory=dict)

    def amend_parts_from_account_day_info(self, account_day_info):
        """
        Called after creating the instance, to populate the parts with remote
        information
        """
        for account_part_info in account_day_info.part_infos.values():
            if account_part_info.part in self.part_infos:
                part_info = self.part_infos[account_part_info.part]
                part_info.has_star = account_part_info.has_star
                part_info.update_status()
            else:
                # noinspection PyTypeChecker
                part_info = CombinedPartInfo(
                    day_info=self,
                    part=account_part_info.part,
                    has_star=account_part_info.has_star,
                    has_code=False,
                    path=None,
                    status=CombinedPartInfo.STATUS_UNKNOWN,
                    module_name='',
                )
                self.part_infos[account_part_info.part] = part_info

    def amend_parts_from_repo_day_info(self, repo_day_info):
        """
        Called after creating the instance, to populate the parts with local
        information
        """
        for repo_part_info in repo_day_info.part_infos.values():
            if repo_part_info.part in self.part_infos:
                part_info = self.part_infos[repo_part_info.part]
                part_info.path = repo_part_info.path
                part_info.has_code = repo_part_info.has_code
                part_info.module_name = repo_part_info.module_name
            else:
                part_info = CombinedPartInfo(
                    day_info=self,
                    part=repo_part_info.part,
                    has_star=False,
                    path=repo_part_info.path,
                    has_code=repo_part_info.has_code,
                    status=CombinedPartInfo.STATUS_UNKNOWN,
                    module_name=repo_part_info.module_name,
                )
                self.part_infos[repo_part_info.part] = part_info
            part_info.update_status()

    @property
    def year(self):
        return self.year_info.year

    @property
    def relative_path(self):
        return self.path.relative_to(settings_proxy().challenges_root)

    @property
    def has_code(self):
        return bool(self.parts_with_code)

    @property
    def has_site_data(self):
        return self.year_info.has_site_data

    def get_input_filename(self):
        return settings_proxy().challenges_boilerplate\
            .get_day_input_filename(self.year, self.day)

    def get_year_url(self) -> str:
        return self.year_info.get_year_url()

    def get_day_url(self) -> str:
        return WebAoc().get_day_url(self.year, self.day)

    def get_part(self, part) -> Optional['CombinedPartInfo']:
        part_info = self.part_infos.get(part)
        if not part_info:
            click.echo(
                f"It looks like there is no code for "
                f"{e_error(f'{self.year} day {self.day} part {part}')}")
            return None
        return part_info

    def serialise(self):
        return {
            "year": self.year,
            "day": self.day,
            "stars": self.stars,
            "parts_with_code": self.parts_with_code,
            "parts": {
                part: part_info.serialise()
                for part, part_info in self.part_infos.items()
            }
        }


@dataclass
class CombinedPartInfo:
    """
    Part information.
    """
    day_info: CombinedDayInfo
    part: str
    has_star: bool
    has_code: bool
    status: str
    path: Path
    module_name: str

    STATUS_UNKNOWN = 'unknown'
    STATUS_COMPLETE = 'complete'
    STATUS_FAILED = 'failed'
    STATUS_DID_NOT_ATTEMPT = 'did-not-attempt'
    STATUS_COULD_NOT_ATTEMPT = 'could-not-attempt'

    STATUSES = [
        STATUS_COMPLETE,
        STATUS_FAILED,
        STATUS_DID_NOT_ATTEMPT,
        STATUS_COULD_NOT_ATTEMPT,
    ]

    def __post_init__(self):
        self.update_status()

    def update_status(self):
        """Update the status, used when we add new information"""
        self.status = self.get_status()

    @property
    def year(self):
        return self.day_info.year

    @property
    def day(self):
        return self.day_info.day

    @property
    def is_final_part(self):
        return (self.day, self.part) == (25, "b")

    @property
    def relative_path(self):
        return self.path.relative_to(settings_proxy().challenges_root)

    @property
    def has_site_data(self):
        return self.day_info.has_site_data

    def get_status(self):
        """Calculate the status, based on the whether it's part A or B"""
        if self.part == 'a':
            return self.get_part_a_status()
        elif self.part == 'b':
            return self.get_part_b_status()
        else:
            raise Exception(f"Unknown part '{self.part}'")

    def get_part_a_status(self):
        """Calculate the status, as if it was part A"""
        if not self.has_site_data:
            return self.STATUS_DID_NOT_ATTEMPT
        elif self.day_info.stars >= 1:
            return self.STATUS_COMPLETE
        elif self.has_code:
            return self.STATUS_FAILED
        else:
            return self.STATUS_DID_NOT_ATTEMPT

    def get_part_b_status(self):
        """Calculate the status, as if it was part B"""
        if not self.has_site_data:
            return self.STATUS_DID_NOT_ATTEMPT
        elif self.day_info.stars == 2:
            return self.STATUS_COMPLETE
        elif self.day_info.stars == 1:
            if self.day == 25 and self.day_info.year_info.stars < 49:
                return self.STATUS_COULD_NOT_ATTEMPT
            elif self.has_code:
                return self.STATUS_FAILED
            else:
                return self.STATUS_DID_NOT_ATTEMPT
        elif self.day_info.part_infos['a'].has_code:
            return self.STATUS_COULD_NOT_ATTEMPT
        else:
            return self.STATUS_DID_NOT_ATTEMPT

    def get_input_filename(self):
        return self.day_info.get_input_filename()

    def get_year_url(self) -> str:
        return self.day_info.get_year_url()

    def get_day_url(self) -> str:
        return self.day_info.get_day_url()

    def get_challenge_instance(self):
        """Get the challenge instance from the module"""
        module = self.get_module()
        if not module:
            return None

        if not hasattr(module, 'challenge'):
            challenge_class = getattr(module, 'Challenge')
            if not isinstance(challenge_class, type) \
                    or not issubclass(challenge_class, BaseChallenge):
                click.echo(
                    f"Challenge {e_error(module.__name__)} does not use "
                    f"`BaseChallenge` and doesn't specify a `challenge` "
                    f"instance")
                return None
            challenge_instance = challenge_class()
        else:
            challenge_instance = getattr(module, 'challenge')

        if not isinstance(challenge_instance, BaseChallenge):
            click.echo(
                f"Challenge {e_error(module.__name__)} `challenge` instance is "
                f"not of `BaseChallenge`")
            return None

        return challenge_instance

    def get_module(self):
        """Load the module for this challenge part"""
        module = try_import_module(self.module_name)
        if not module:
            click.echo(f"Could not find {e_error(self.module_name)}")

        return module

    def serialise(self):
        return {
            "year": self.year,
            "day": self.day,
            "part": self.part,
            "has_star": self.has_star,
            "has_code": self.has_code,
            "status": self.status,
            "module_name": self.module_name,
        }



from dataclasses import dataclass, field
from typing import Optional

import bs4
import click
import requests

from aox.settings import settings_proxy
from aox.styling.shortcuts import e_error


__all__ = ['WebAoc']


@dataclass
class WebAoc:
    """
    Helper class that abstracts requests to the AOC site.

    The session ID is necessary before any request.
    """
    session_id: Optional[str] = field(
        default_factory=lambda: settings_proxy().aoc_session_id)

    root_url = 'https://adventofcode.com'
    headers = {
        "User-Agent": "aox",
    }
    cookies = {}

    def get_events_url(self):
        """
        >>> WebAoc('test-session').get_events_url()
        'https://adventofcode.com/events'
        """
        return f"{self.root_url}/events"

    def get_year_url(self, year):
        """
        >>> WebAoc('test-session').get_year_url(2020)
        'https://adventofcode.com/2020'
        """
        return f"{self.root_url}/{year}"

    def get_day_url(self, year, day):
        """
        >>> WebAoc('test-session').get_day_url(2020, 5)
        'https://adventofcode.com/2020/day/5'
        """
        return f"{self.root_url}/{year}/day/{day}"

    def get_input_url(self, year, day):
        """
        >>> WebAoc('test-session').get_input_url(2020, 5)
        'https://adventofcode.com/2020/day/5/input'
        """
        return f"{self.root_url}/{year}/day/{day}/input"

    def get_answer_url(self, year, day):
        """
        >>> WebAoc('test-session').get_answer_url(2020, 5)
        'https://adventofcode.com/2020/day/5/answer'
        """
        return f"{self.root_url}/{year}/day/{day}/answer"

    def is_configured(self):
        """
        >>> WebAoc('test-session').is_configured()
        True
        >>> WebAoc(None).is_configured()
        False
        >>> WebAoc('').is_configured()
        False
        """
        return bool(self.session_id)

    def get_events_page(self):
        """Get the page with stars per year"""
        return self.get_html(self.get_events_url(), 'events information')

    def get_year_page(self, year):
        """Get the page with stars per day"""
        return self.get_html(
            self.get_year_url(year), f"year {year} information")

    def get_input_page(self, year, day):
        """Get the input for a particular day"""
        return self.get_text(
            self.get_input_url(year, day), f"year {year} day {day} input")

    def submit_solution(self, year, day, part, solution):
        """Post a solution"""
        return self.post_html(
            self.get_answer_url(year, day), {
                "level": 1 if part == "a" else 2,
                "answer": solution,
            }, f"year {year} day {day} input")

    def get_html(self, url, parse_name, *args, **kwargs):
        """Get parsed HTML"""
        return self.get(
            url=url, parse_type='html', parse_name=parse_name,
            *args, **kwargs)

    def get_text(self, url, parse_name, *args, **kwargs):
        """Get raw text"""
        return self.get(
            url=url, parse_type='text', parse_name=parse_name,
            *args, **kwargs)

    def get(self, *args, **kwargs):
        """Get a page"""
        return self.fetch(requests.get, *args, **kwargs)

    def post(self, *args, **kwargs):
        """Submit a request"""
        return self.fetch(requests.post, *args, **kwargs)

    def post_html(self, url, data, parse_name, *args, **kwargs):
        """Post and return parsed HTML"""
        return self.post(
            url=url, data=data, parse_type='html', parse_name=parse_name,
            *args, **kwargs)

    def fetch(self, method, url, extra_headers=None, extra_cookies=None,
              parse_type=None, parse_name=None, data=None):
        """
        Generic method for interacting with the site. It can also parse the the
        result as a particular type (eg HTML).
        """
        if not self.is_configured():
            return None

        if not url.startswith(self.root_url):
            raise Exception(
                f"Only AOC URLs can be accessed (starting with "
                f"'{self.root_url}'), not '{url}'")

        response = method(
            url,
            data=data,
            headers=self.get_headers(extra_headers),
            cookies=self.get_cookies(extra_cookies),
        )

        if parse_type:
            return self.parse(response, parse_type, parse_name)

        return response

    def get_headers(self, extra_headers=None):
        """
        Construct the headers for a request

        >>> WebAoc('test-session').get_headers()
        {'User-Agent': 'aox'}
        >>> WebAoc('test-session').get_headers({})
        {'User-Agent': 'aox'}
        >>> WebAoc('test-session').get_headers({'User-Agent': 'test'})
        {'User-Agent': 'test'}
        >>> WebAoc('test-session').get_headers(
        ...     {'User-Agent': 'test', 'foo': 'bar'})
        {'User-Agent': 'test', 'foo': 'bar'}
        """
        return {
            **self.headers,
            **(extra_headers or {}),
        }

    def get_cookies(self, extra_cookies=None):
        """
        Construct the cookies for a request

        >>> WebAoc('test-session').get_cookies()
        {'session': 'test-session'}
        >>> WebAoc('test-session').get_cookies({})
        {'session': 'test-session'}
        >>> WebAoc('test-session').get_cookies({'session': 'other'})
        {'session': 'other'}
        >>> WebAoc('test-session').get_cookies(
        ...     {'session': 'other', 'foo': 'bar'})
        {'session': 'other', 'foo': 'bar'}
        """
        return {
            "session": self.session_id,
            **self.cookies,
            **(extra_cookies or {}),
        }

    def parse(self, response, _type, name):
        """
        Parse a response as a particular type (eg HTML)

        >>> from requests import Response
        >>> _response = Response()
        >>> _response._content = b'<html><body><article>Hi'
        >>> _response.status_code = 200
        >>> html = WebAoc('test-session').parse(_response, 'html', 'test')
        >>> html
        <html><body><article>Hi</article></body></html>
        >>> html.article
        <article>Hi</article>
        >>> _response = Response()
        >>> _response._content = b'Hello there'
        >>> _response.status_code = 200
        >>> WebAoc('test-session').parse(_response, 'text', 'test')
        'Hello there'
        """
        if _type == 'html':
            return self.as_html(response, name)
        if _type == 'text':
            return self.as_text(response, name)
        else:
            raise Exception(f"Unknown parse type '{_type}'")

    def as_html(self, response, name):
        """
        Parse a response as HTML

        >>> from requests import Response
        >>> _response = Response()
        >>> _response._content = b'<html><body><article>Hi'
        >>> _response.status_code = 200
        >>> html = WebAoc('test-session').as_html(_response, 'test')
        >>> html
        <html><body><article>Hi</article></body></html>
        >>> html.article
        <article>Hi</article>
        >>> _response = Response()
        >>> _response._content = b'Oops'
        >>> _response.status_code = 400
        >>> WebAoc('test-session').as_html(_response, 'test')
        """
        if not response:
            return None

        if not response.ok:
            click.echo(
                f"Could not get {e_error(name)} from the AOC site "
                f"({response.status_code}) - is the internet down, AOC down, "
                f"the URL is wrong, or are you banned?")
            return None

        return bs4.BeautifulSoup(response.text, "html.parser")

    def as_text(self, response, name):
        """
        Parse a response as text

        >>> from requests import Response
        >>> _response = Response()
        >>> _response._content = b'Hello there'
        >>> _response.status_code = 200
        >>> WebAoc('test-session').as_text(_response, 'test')
        'Hello there'
        >>> _response = Response()
        >>> _response._content = b'Oops'
        >>> _response.status_code = 400
        >>> WebAoc('test-session').as_text(_response, 'test')
        """
        if not response:
            return None

        if not response.ok:
            click.echo(
                f"Could not get {e_error(name)} from the AOC site "
                f"({response.status_code}) - is the internet down, AOC down, "
                f"the URL is wrong, or are you banned?")
            return None

        return response.text
